/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


/*
 * slice_manager.h
 */
#include "replication_policy.h"
#include "wfs_rsm_cclient.h"
#include "wfs_file.h"
#include "wfs_slice.h"
#include "wcc_protocol.h"
#include <wfs_client_manager.h>
#include "rpc.h"
#include <vector>
#include <map>
#include <string>

#ifndef SLICE_MANAGER_H_
#define SLICE_MANAGER_H_

class slice_manager {
  public:
    enum ret_code {
      SM_CANT_REPAIR = -5,
      SM_NOENT = -4,
      SM_RSM_ERR = -3,
      SM_CANT_SATISFY = -2,  // Can't satisfy policy at all (Don't know the requested primary site).
      SM_GENERAL_ERR = -1,
      SM_SUCCESS = 0,
      SM_TRYHOST = 1,  // Requested policy whose a known host is located at the primary site (but not this host)
      SM_NOLOCKED = 2,
      SM_SA_INC = 3,  // Can't create a slice assignment that fully satisfies the policy.
    };
  private:
    struct slice_ent {
      // Should there be an attempt to correct the assignment.
      enum sat_state {
        OK,
        REPAIR,
        SYNCING,
      };
      // Possible ways not that an assignment does not comply with the policy
      enum sa_problem {
        NOT_ENOUGH_HOSTS = ((unsigned int)1) << 1,
        NOT_ENOUGH_SITES = ((unsigned int)1) << 2,
        WRONG_PRIMARY_SITE = ((unsigned int)1) << 3,
      };

      static const uint64_t CAPACITY = ((uint64_t) ~wfs_slice::SLICE_MASK) + 1;

      slice_assignment sa;
      //wfs_file_id last_used_id;
      std::string lock_state; // This is just for finding out faster if the lock has changed.
      bool locked;
      /*
       * The count is kept for internal uses only and is not replicated. It is used only to know
       * How many fid were assigned by this host since it has been up.
       */
      unsigned int count;
      sat_state compliance;
      unsigned int repair_items;
      std::set<host> old_replicas; // Hosts that at some earlier time were replicas for this slice.

    slice_ent(std::string state) : sa(state) , lock_state(""), locked(false), count(0), compliance(OK)
        , repair_items(0){}
      slice_ent(wfs_slice sl, rep_policy pol, std::list<host> lst)
      : sa(sl,pol,lst), lock_state(""), locked(false), count(0), compliance(REPAIR), repair_items(0) {}

      host primary() const {if (sa.replicas.size() > 0) return sa.get_primary(); else return host();}
      bool amiprimary(host me) const {return (primary() == me);}	// Does not check site.

      // If this slice is held by this host, check with rsm client O/W Just look at
      // what was last reported by the cserver.
      bool islocked(wfs_rsm_cclient * ccl,host me) const {
        if(amiprimary(me))
          return ccl->is_slice_locked(sa.slice.val());
        else return locked;
      }
      wfs_file_id get_fid(); // Get a random file id in this slice.
      // How many files were assigned under this slice
      //unsigned int fid_count() const { return (~wfs_slice::SLICE_MASK & last_used_id)+1; }
      unsigned int fid_count() const { return count;}
      /*
       * Are all file id's for this slice assigned.
       * Since we choose the file ID's from a slice randomly and get rejected by the caller if this ID exists,
       * When many slices have been given, the chances of success are slim. So, when a slice is at 8/10 capacity,
       * mark it as full - this will cause an expected 8 calls by the caller. and by Chebyshev's, prob(more than 30)
       * is less than 0.05
       */
      bool isfull() { return fid_count() >= (CAPACITY*0.8);}
      // Need this for sorting some lists. We sort by the number of given file ids
      bool operator <(const slice_ent & other) const {return fid_count() < other.fid_count();}

      bool is_errstat(sa_problem prob) {return ((repair_items & prob)!= 0);}
      void set_errstat(sa_problem prob) {repair_items |= prob;}
      void reset_errstat(sa_problem prob) {repair_items &= ~prob;}
      void reset_all_errstats() { repair_items = 0;}
    };

    host _me;
    // This should be either taken from a dynamic DNS server or a pre-configured list of masters.
    std::string _master;
    wfs_rsm_cclient * _ccl;
    unsigned int _last_sync; // When did we last sync with the server
    bool _try_root_slice;
    int _sync_period;
    viewstamp _update_stamp;
    bool _shutting_down;
    int _lease_period;

    typedef std::map<wfs_slice_id,slice_ent *>  slicemap;
    slicemap _slices;
    mutable pthread_mutex_t _slices_mutex;

    // Map sites to a set of hosts. (This host is not in the set).
    typedef std::map<std::string,std::set<host> > sitemap;
    sitemap _sites;
    mutable pthread_mutex_t _sites_mutex;

    // Map policies as string keys to slices held by this host satisfying these policies.
    // The lists are sorted by the number of assigned files for an entry
    // protected by slice_mutex
    typedef std::map<std::string,std::set<slice_ent *> > policymap;
    policymap _sat_policies;
    static const std::string make_key_from_pol(const rep_policy& pol) {
      rep_policy pol2(pol);
      return static_cast<std::string>(pol2);}

    /*
     * Assigning random unused slice IDs:
     * We keep a set of slices known to be in use. When asked for a new slice id we randomly choose
     * a slice ID between 0-MAX_SLICE_ID until we draw a number that hasn't been assigned. This
     * Should be efficient enough for a small set of used slices.
     * Optimization: Once the number of slices reaches MAX_SLICE_COUNT/2 we generate an array of
     * all slices. We then use the set of already assigned slice ID's as the first random draws
     * in a Knuth shuffle and from then on we just choose a number without repetition using the array
     * and a Knuth shuffle.
     */
    unsigned int _usedslices_count;
    std::vector<wfs_slice_id> _slices_perm;

    // PRIVATE AUXILLIARY CALLS
    // Got to have _slices_mutex locked to call find_unused_slice
    ret_code find_unused_slice(wfs_slice_id & sid);
    host get_host_in_site(std::string site);
    bool get_slices_updates();
    friend bool compare_by_fid_count(slice_ent * pFirst,slice_ent * pSecond);
    // Create a new slice assignment that satisfies the policy.Lock the new slice.
    ret_code create_new_slice_entry(const rep_policy& pol, slice_ent ** ppEnt,
				    wfs_slice_id forced_id = 1, 
				    std::vector<host> forced_replicas = std::vector<host>() );
    // Use the exact same slice assignment with a new slice. (Lock the new slice).
    ret_code duplicate_slice_assignment(const slice_ent * copyfrom, slice_ent ** ppEnt);
    // Check id the assignment needs to be corrected and mark it that way.
    bool does_entry_comply(slice_ent * pEnt, bool dontCheckWhenSyncing = true);
    // Try and correct the entry. Add hosts to be recruited. Mark entry as being synced.
    ret_code suggest_repair(slice_ent * pEnt,std::list<std::pair<wfs_slice_id,host> > & recruits,
        const std::set<host> active_set);
    void slice_update(slice_ent * pEnt, bool released);
    slice_manager::ret_code get_active_set(std::set<host> & accum_set);

    inline void find_hosts_in_site(const wfs_slice_id sidx, const std::string sitename,
        int & needed_hostsx,std::set<host> hostsetx,
        std::list<std::pair<wfs_slice_id,host> >& recruitsx,
        slice_manager::ret_code& retx);

    ret_code
    get_random_replica(int i, int rep_sites, std::set<std::string>& sites_used,const std::vector<std::string>& sites_vec,
        std::set<host> nodes_used ,std::map<std::string,std::pair<int,int> >& nodes_per_site, host & h, std::string& s);

    /*
     * Implementation note: single slice_ent are not protected by a mutex. both
     * _slices_mutex and _pols_mutex access entries. _pols_mutex does only write operation
     * which is get_fid(). Since the side effects of this operation are not very important for the
     * operation, I think it is not worth adding a mutex. maybe make  _pols_mutex hold const pointers. Y.S.
     */
  public:

    static const unsigned int DEFAULT_LEASE_PERIOD = 20; // 20 Seconds
    explicit slice_manager(bool from_log, std::string dst, host me,
			   wfs_client_manager *clients,
			   int sync_period = 10,
			   int lease_period = DEFAULT_LEASE_PERIOD);
    ~slice_manager();

    void loop();

    // Get all the known slices and assignments from the cserver TODO:Make private
    bool sync_state();

    /*
     * Check to see if this fid falls within a slice that this host is registered (locked) as a primary for.
     * If the manager had no knowledge of this slice the answer is false.
     */
    bool is_primary_for_file(const wfs_file_id fid) const;

    /*
     * Given a replication policy pol try and get a fid that falls within the slice that satisfies this policy
     * 	- First, check to see if the site for the primary is the local site. If not look for a random
     * 		host in the requested site, put in tryhost and return SM_TRYHOST.
     * 	- Else, See if there is an existing slice locked by this host that satisfies the policy, if so get the next
     * 		available fid from that slice.
     * 	- Else, Find a new unlocked slice. Create a replica set for it, lock it and return the first fid from that
     * 		slice.
     *  - A slice will be created on a best effort method. If the assignment can't fully comply with the policy
     *    the slice is still created and locked,SM_SA_INC will be returned,
     *    and that assignment will be fixed in the future opportunistically.
     *  Setting force_create, causes a new slice to be created even if there is a slice that complies with pol.
     *
     * return:
     *  SM_SUCCESS: A valid assignment has been locked (or exists).
     *  SM_SA_INC: See above.
     */
    slice_manager::ret_code get_new_fid(const rep_policy& pol,wfs_file_id& fid, host& tryhost,
					bool force_local_primary = false,
					wfs_slice_id forced_id = 1,
					std::vector<host> forced_replicas = std::vector<host>());

    // Get all slices for which this host is either a primary or a backup.
    slice_manager::ret_code get_maintenence_slices_for_host(const host h,std::vector<wfs_slice_id>& slicev);
    slice_manager::ret_code get_maintenence_slices(std::vector<wfs_slice_id>& slicev)
      {return get_maintenence_slices_for_host(_me,slicev);}

    /*
     * Get the primary for a slice.
     * If onlyIfLocked is false - then a primary will be returned for EVERY slice this manager knows of.
     * If onlyIfLocked is set, then return the primary ONLY if the manager thinks this slice is locked (by any node).
     * returns:
     * 	SM_SUCCESS - If slice was found and (onlyIfLocked==false || (onlyIfLocked==true && slice is locked))
     * 	SM_NOENT - Slice is not known
     * 	SM_NOLOCKED - onlyIfLocked==true && slice is found && slice is not locked.
     *
     */
    slice_manager::ret_code get_slice_primary(const wfs_slice_id sid, host & prim, bool onlyIfLocked = true);

    /*
     * Get the vector of nodes that are in that slice's replica set.
     * Note that the size of the vector may be larger than the replication number.
     * hosts 0 to (rep_num - 1), should be treated as active replicas, the rest may be old replicas and are kept
     * for helping to choose nodes to recover.
     * (Currently, passive replicas are not so rep_num should equal repv.size() after the call)
     */
    slice_manager::ret_code get_slice_replicaset(const wfs_slice_id sid, std::vector<host>& repv, int & rep_num);
    /*
     *  Return a vector of all active nodes in the system, Those are nodes that are either
     *  primaries or backups for a slice
     */
    slice_manager::ret_code get_all_active_hosts(std::vector<host> & hosts);
    /*
     * Checks all slices that this host is a primary for and sees if they satisfy their corresponding policies.
     * Return a map of wfs_slice_id to hosts to be interpreted as these hosts should be recruited as backups for these
     * slices.
     * Side effect: Each lock entry that can be repaired will have "compliance" changed to sync
     */
    slice_manager::ret_code check_all_slices(std::list<std::pair<wfs_slice_id,host> > & recruits);

    /*
     * Used once the specified host has is ready to assume its role as a back-up.
     * The manager will attempt to lock the new state with the cserver.
     * If the repair included dismissing some hosts for the slice, they will be put in "dropped"
     * If the host can assume the role of a primary to a slice with the wrong primary then
     *  - The slice is released
     *  - do_lock_rpc is set to true.
     *  o/w do_lock_rpc is set to false.
     * return:
     *  SM_SUCCESS: if repaired state has been locked with the cserver.
     */
    slice_manager::ret_code report_sync_complete(const wfs_slice_id sid,const host h,std::vector<host>& dropped);
    slice_manager::ret_code report_sync_complete(const wfs_slice_id sid,const host h,std::vector<host>& dropped,
        bool& do_lock_rpc);

    /*
     * Notify the manager that a sync failed. This allows the manager to try and repair this slice again when
     * check_all_slices is called.
     * Between a first call to check_all_slices that marks an entry for repair
     *  and either report_sync_complete or sync_aborted, repeated calls to check_all_slices will ignore this entry.
     */
    slice_manager::ret_code sync_aborted(const wfs_slice_id sid,const host h);

    /*
     * As the name suggests.
     * Since this is commonly used for slices created on this host. No cserver sync is done if the slice
     * is not found.
     */
    slice_manager::ret_code get_policy_for_slice(const wfs_slice_id sid,rep_policy& pol);

    /*
     * Tries to repair slice assignments that are affected by a down node.
     * Usually the repair process will involve locking a new assignment for an existing slice.
     * host h could have one of the two relations with the caller node:
     * 	- h is a secondary backup replica for a slice whose primary is the caller: h will be removed from that assignment.
     *  - h is a primary for a slice whose replica set include the caller: the caller will promote itself to the primary.
     *  TODO: Maybe the caller should be allowed to be any of the slice's non-primary replicas - to allow recovery when more than one node fails.
     * returns:
     * 	SM_SUCCESS - A new assignment has been locked with the cserver that does not include the downed host.
     *  SM_RSM_ERR - Can't lock the modified slice.
     *  SM_CANT_REPAIR - The slice information obtained from the server suggests that h is still a primary of an
     *     active slice. This suggests that either h is not down or the lease period on that lock has not expired.
     *
     */
    slice_manager::ret_code report_host_down(const host h);

    slice_manager::ret_code release_slice(const wfs_slice_id sid);



    /*
     * Debugging
     */
    void print_all_known_slices();


};


#endif /* SLICE_MANAGER_H_ */
