/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include <arpa/inet.h>
#include <getopt.h>
#include <assert.h>
#include <netdb.h>
#include <rpc.h>
#include <vivaldi.h>
#include <wfs_misc.h>
#include <assert.h>

#include "host_marshall.h"
#include "wfs_cent_cserver.h"


static vivaldi* viv;

wfs_cent_cserver::wfs_cent_cserver( int timeout ) :
  _timeout(timeout) {
  assert( pthread_mutex_init( &_hosts_lock, NULL ) == 0 );
}

wfs_cent_cserver::~wfs_cent_cserver() {
  pthread_mutex_destroy( &_hosts_lock );
}

int wfs_cent_cserver::ping( host h, host &dummy ) {


  wfs_log(WFS_DBG_3, "Got a ping from %s:%d_%s\n", inet_ntoa( h.sin().sin_addr ),
	    htons(h.port), h.site_id.c_str() );


  pthread_mutex_lock( &_hosts_lock );

  //return a random node I know to the ping requester
  if( _hosts.size() > 0 ) {
    int r = random() % _hosts.size();
    for( unsigned i = 0; i < _hosts.size(); i++ ) {
      int x = (r+i) % _hosts.size();
      if( _hosts[x] != h ) {
	dummy = _hosts[x];
	break;
      }
    }
  }

  // update the time for this host in _hosts
  for( uint i = 0; i < _hosts.size(); i++ ) {
    if( _hosts[i] == h ) {
      _hosts[i].last = time(0);
      pthread_mutex_unlock( &_hosts_lock );
      return wcc_protocol::OK;
    }
  }

  host1 new_host( h, time(0) );
  _hosts.push_back( new_host );
  pthread_mutex_unlock( &_hosts_lock );
  return wcc_protocol::OK;

}

int wfs_cent_cserver::fetch( int dummy, std::vector<host> &hosts ) {

  pthread_mutex_lock( &_hosts_lock );

  for( uint i = 0; i < _hosts.size(); i++ ) {
    if( _hosts[i].up( _timeout ) ) {
      hosts.push_back( _hosts[i] );
    }
  }

  pthread_mutex_unlock( &_hosts_lock );

  return wcc_protocol::OK;

}


static void usage() {

  fprintf( stderr, "Usage for wfs_cent_cserver:\n\n"
	   "\t--port <port>\n"
	   "\t--debug (may be given multiple times)\n"
	   "\t--timeout <timeout_secs>\n"
#ifdef WFS_WITH_LIBSSH
	   "\t[--secure]\n"
	   "\t[--keydir <path>] (defaults to ~/.wfs)\n"
#endif
	   );

  exit(-1);

}

int main( int argc, char *argv[] ) {

  srandom(time(NULL));

  unsigned int port = 0;
  int debug_level = 0;
  bool secure = false;
  const char *key_dir = default_keydir;

  int c;
  optind = 0;  // reset getopt
  int timeout = 60; // in seconds
  //int timeout = 180; // in seconds

  while(1) {

    static struct option long_options[] =
    {
      {"port", required_argument, 0, 'p'},
      {"timeout", required_argument, 0, 't'},
      {"debug", no_argument, 0, 'd'},
#ifdef WFS_WITH_LIBSSH
      {"secure", no_argument, 0, 'S'},
      {"keydir", required_argument, 0, 'K'},
#endif
      {"help", no_argument, 0, 'h'},
      {0, 0, 0, 0}
    };

    int option_index = 0;
    c = getopt_long( argc, argv, "p:dt:Sh?",
		     long_options, &option_index );

    if( c == -1 ) {
      break;
    }

    switch(c) {

    case 'p':
      port = atoi(optarg);
      break;

    case 't':
      timeout = atoi(optarg);
      assert( timeout > 0 );
      break;

    case 'd':
      debug_level++;
      break;

#ifdef WFS_WITH_LIBSSH
    case 'S':
      secure = true;
      break;
    case 'K':
      key_dir = optarg;
      break;
#endif

    case '?':
    case 'h':
      /* getopt_long already printed an error message. */
      usage();
      break;

    default:
      usage();
    }
  }

  if( debug_level > 0 ) {
    wfs_set_debug( debug_level );
    printf( "DEBUG LEVEL: %d\n", debug_level );
  }

  if (secure) {
    rpc_enable_security(wfs_public_key_identifies_user,
			wfs_get_private_key,
			"host");
    wfs_set_key_dir(key_dir);
  }

  wfs_thread_id_init();

  if( port == 0 ) {
    fprintf( stderr, "Required parameter is missing.\n" );
    usage();
  }

  fflush(stdout);
  setlinebuf(stdout);

  rpcs server( htons(port) );
  //server.setlossy(true);
  viv = new vivaldi();
  server.set_vivaldi(viv);
  wfs_cent_cserver cserver( timeout );

  server.reg(wcc_protocol::ping, &cserver, &wfs_cent_cserver::ping);
  server.reg(wcc_protocol::fetch, &cserver, &wfs_cent_cserver::fetch);

  while(1)
    sleep(1000);
}
