/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


// this is the file server
// wfs_client has a similar interface
// overserve.cc turns received RPCs (from wfs_client) into calls on wfs_server

#ifndef __WFS_SERVER_H__
#define __WFS_SERVER_H__ 1

#include "wfs_protocol.h"

#include <host.h>
#include <map>
#include <string>
#include <vector>
#include <slice_manager.h>
#include <wfs_client_manager.h>
#include <wfs_lease_manager.h>
#include <wfs_maint.h>
#include <wfs_perfile_locker.h>
#include <wfs_metadata.h>
#include "wfs_cacheloc.h"
#include "wfs_file.h"
#include <wfs_metadata.h>
#include <wfs_storage.h>

bool find_fid( FILE *fp, const std::string &name, char *line, bool use_line, 
	       wfs_file_id &fid, bool &higher );

class wfs_server {
 
  static wfs_file_handle empty_fh;

 private:
  std::string _dir;
  wfs_cacheloc _cache_loc;
  wfs_perfile_locker *_locker;
  wfs_maint *_maint;
  wfs_storage *_storage;
  wfs_client_manager *_clients;
  slice_manager *_slices;
  wfs_lease_manager _leases;
  std::map< wfs_node_id, std::map< wfs_file_id, bool > > _invalidate_sleepers;
  pthread_mutex_t _is_lock;
  int _num_threads;
  pthread_mutex_t _nt_lock;
  pthread_cond_t _nt_cond;


  std::string cue_filename(wfs_file_handle);
  std::string version_filename(wfs_file_handle);
  std::string read_perm_cues( wfs_file_handle fh );
  wfs_version read_version( wfs_file_handle fh );
  const void save_old_version( wfs_file_handle fh );
  const wfs_version update_version( wfs_file_handle fh );
  const wfs_version update_version( wfs_file_handle fh, wfs_metadata *md );
  void call_invalidate( wfs_file_handle fh, wfs_lease_manager::leaser leaser );
  void invalidate_leases( wfs_file_handle fh, host modifier, 
			  wfs_file_handle other_fh = empty_fh );
  int icreate_dirent( wfs_file_handle dh, std::string name,
		      wfs_file_handle new_fh, host client,
		      wfs_file_handle other_fh = empty_fh );
  int iunlink( wfs_file_handle dh, std::string name, host client,
	       wfs_file_handle other_fh = empty_fh, bool sync = true );
  int icreate_file( wfs_file_handle,
		    const std::map<std::string,wfs_permission> &permissions );
  int do_write( wfs_file_handle, std::vector<unsigned long long> off, 
		std::vector<std::string>, const std::string &user,
		wfs_version &new_version);
  int do_truncate( wfs_file_handle, unsigned long long off, 
		   const std::string &user,
		   wfs_version &new_version );
  int invalidate_and_sync( wfs_file_handle fh, host modifier, 
			   wfs_file_handle other_fh = empty_fh,
			   std::string append_data = "" );
  bool check_stale( wfs_file_handle fh );
  bool can_handle( wfs_file_handle fh );
  void new_id( wfs_file_handle &new_fh, uint16_t type, 
	       wfs_file_handle parent, bool force_local = false );

 public:
  wfs_server( std::string dir, wfs_perfile_locker *locker, 
	      wfs_maint *maint, wfs_storage *storage, 
	      wfs_client_manager *clients, slice_manager *slices );
  ~wfs_server();

  bool isfile( wfs_file_handle );
  bool isdir( wfs_file_handle );
  wfs_file_handle ilookup( wfs_file_handle di, std::string name );

  int pingresponder( host h, host &dummy );
  int getfile( const std::string &user, wfs_file_handle, host client,
	       wfs_protocol::fileinfo & );
  int getfiles( const std::string &user, std::vector<wfs_file_handle>, host client,
		std::vector<wfs_protocol::fileinfo> & );
  int getdir( const std::string &user, wfs_file_handle, host client, 
	      wfs_protocol::dirinfo & );
  int readdir( const std::string &user, wfs_file_handle, 
	       unsigned long long ind, unsigned int n, host client, 
	       std::vector<wfs_storage::dirent> & );
  int readfile( const std::string &user, wfs_file_handle, 
		wfs_version version, unsigned long long off, 
		unsigned int n, wfs_protocol::readinfo & );
  int writefile( const std::string &user, wfs_file_handle, 
		 unsigned long long off, std::string, 
		 host client, wfs_version &new_version );
  int writefile_batch( const std::string &user, wfs_file_handle, 
		       std::vector<wfs_protocol::write>, 
		       host client, wfs_version & );
  int lookup( const std::string &user, wfs_file_handle, std::string name, 
	      host client, wfs_protocol::lookupinfo & );
  int rename( const std::string &user, wfs_file_handle old_parent, 
	      std::string old_name, wfs_file_handle new_parent, 
	      std::string new_name, host client, int &dummy );
  int create_dirent( const std::string &user, 
		     wfs_file_handle, std::string name, wfs_file_handle, 
		     wfs_protocol::create_file_struct, 
		     host client, wfs_protocol::file_and_version &final_file );
  int unlink( const std::string &user, wfs_file_handle, std::string name, 
	      int is_dir, int dec_link, host client, 
	      wfs_protocol::file_and_version & );
  int create_file( const std::string &user, wfs_file_handle, uint16_t type,
		   std::map<std::string,wfs_permission> permissions, 
		   wfs_file_handle &final_fh );
  int truncate( const std::string &user, wfs_file_handle, 
		unsigned long long off, host client, wfs_version &new_version );
  int chmod( const std::string &user, wfs_file_handle, 
	     std::map<std::string,wfs_permission>, 
	     host client, wfs_version &new_version);
  int add_link( const std::string &user, wfs_file_handle fh, int num_add,
		host client, wfs_version &new_version );

  int addcacheloc( wfs_file_handle, wfs_version, cacheloc_ent, int &);
  int getcacheloc( wfs_file_handle, wfs_protocol::blockinfo, Coord c, int, 
		   std::vector<cacheloc_ent>&);
  int renew_lease_batch( std::vector<wfs_file_handle>, host client, int& );

};

#endif // __WFS_SERVER_H__
