/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
#include <unistd.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <string.h>
#include <assert.h>
#include <stdarg.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "wfs_log.h"

#define MAX_FILE_DIGITS 2 // Max number of digits to suffix to a log file

/*
 * This is the main logging class, shared by users of this library.
 * It will be created by WFSLogInit.
 */
WFSLog * _wfslog = 0;

/*
 * The Global Debugging current state.
 */
int WFS_DEBUG_LEVEL = 0;

unsigned int WFSLogInit::_refcount = 0;

/*
 * Set the global debug level to a new level
 */
void
wfs_set_debug(int level) {
  WFS_DEBUG_LEVEL = level;
}

/*
 * For the time being, we use generic log levels
 */
static const char* db_levels[WFS_DBGLEVELS+1] = {"WHEELFS","LEVEL1","LEVEL2","LEVEL3","LEVEL4","EXT"};

WFSLogInit::WFSLogInit(std::string base_name,
                       std::string log_dir,
                       int max_size,
                       int max_files,
                       WFSLogAugmentor * aug)
{
  if(_refcount++ == 0) {
    _wfslog = new WFSLog(base_name,log_dir,max_size,max_files);
    if(_wfslog && aug) {
      _wfslog->addAugmentor(aug);
    }
    _wfslog->redirectAll();
  }
}

WFSLogInit::~WFSLogInit()
{
  // Close files
  if(--_refcount == 0){
    delete _wfslog;
    _wfslog = NULL;
  }
}

/*
 * max_size = Maximum log file size in MB. Not a hard limit. Will try to be close to that limit.
 * max_files = Number of files to roll
 */
WFSLog::WFSLog(std::string base_name, std::string log_dir, int max_size, int max_files)
  : _max_size_b(max_size*(1<<20)), _max_files(max_files), _cur_size(0), _use_stderr(false),
    _augmentor(NULL), _oldstdout(0), _oldstderr(0), _listener(0), _other_socket(0)
{
  assert(pthread_mutex_init(&_ro_lock, NULL) == 0);
  if (_max_files > 100) {
    _max_files = 100;
  }
  if (_max_files < 0 ) {
    _max_files = 2;
  }
  _base_path = log_dir + '/' + base_name + ".log";
  _logfile = fopen(_base_path.c_str(),"a+");
  if (!_logfile) {
      perror("Cannot open log stream. Printing to stderr");
      _logfile = stderr;
      _use_stderr = true;
  }
  // File may have existed
  if (!_use_stderr) {
    _cur_size = lseek(fileno(_logfile), 0, SEEK_END);
  }
  // Write some preamble
  timeval tv;
  gettimeofday(&tv, NULL);
  _cur_size += fprintf(_logfile,"*******************************************\n");
  _cur_size += fprintf(_logfile,"* %d.%06d: Start of Log Segment *\n", int (tv.tv_sec), int (tv.tv_usec) );
  _cur_size += fprintf(_logfile,"*******************************************\n");

}

WFSLog::~WFSLog()
{
  if( _listener != 0 ) {
    assert(pthread_cancel(_listener) == 0);
    assert(pthread_join(_listener, NULL) == 0);
    fclose(_local_socket);
    close(_other_socket);
  }
  pthread_mutex_destroy(&_ro_lock);
  if (_logfile && !_use_stderr) {
    fclose(_logfile);
  };
  if(_augmentor) {
    delete _augmentor;
  }
  _augmentor = NULL;
  // redirect back
  if( _oldstdout != 0 ) {
    close(1);
    dup(_oldstdout);
    close(_oldstdout);
    close(2);
    dup(_oldstderr);
    close(_oldstderr);
  }
}

/*
 * Assumption: Lock is held
 */
bool
WFSLog::do_log_rollover()
{
  if (_cur_size <= _max_size_b){
     //Some other thread beat me to it.
     return true;
  }
  fflush(_logfile);
  fclose(_logfile);

  if (_max_files == 1){
    // Just truncate the file and return
    _logfile = fopen(_base_path.c_str(),"w+");
    _cur_size = 0;
    return true;
  }
  char frombuf[MAX_FILE_DIGITS+1];
  char tobuf[MAX_FILE_DIGITS+1];
  // Rotate the logs
  for (int i = _max_files - 2; i >= 0 ; --i){
    std::string frompath;
    if(i > 0 ){
      snprintf(frombuf,MAX_FILE_DIGITS+1,"%d\0",i-1);
      frombuf[MAX_FILE_DIGITS] = '\0';
      frompath = _base_path + "." + std::string(frombuf);
      if(access(frompath.c_str(), F_OK) == -1 && errno == ENOENT){
          continue;
      }
    }
    snprintf(tobuf,MAX_FILE_DIGITS+1,"%d\0",i);
    tobuf[MAX_FILE_DIGITS] = '\0';
    std::string topath = _base_path + "." + std::string(tobuf);
    if(i > 0){
      rename(frompath.c_str(), topath.c_str()); // Can't do much about failures. might lose some logs
    } else {
      rename(_base_path.c_str(), topath.c_str());
    }
  }
  _logfile = fopen(_base_path.c_str(),"w+");
  _cur_size = 0;
  return true;
}

void cleanup_lock( void *arg ) {
  pthread_mutex_t *lock = (pthread_mutex_t *) arg;
  pthread_mutex_unlock(lock);
}

void
WFSLog::wfslogx(int level,const char* fmt,...)
{
  va_list args;

  if (abs(level) < WFS_DBMIN_LEVEL) {
    level = WFS_DBMIN_LEVEL * (level<0)?-1:1;
  }
  if (abs(level) > WFS_DBMAX_LEVEL) {
    // highest one that's not external
    level = (WFS_DBMAX_LEVEL-1) * (level<0)?-1:1;
  }

  timeval _tv;
  gettimeofday(&_tv, NULL);
  // I wish we could avoid this locking. Only needed for log rollover.
  assert(pthread_mutex_lock(&_ro_lock)==0);
  pthread_cleanup_push( &cleanup_lock, &_ro_lock );
  if(level >= 0) {
    _cur_size += fprintf(_logfile, "%d.%06d: ", int (_tv.tv_sec), 
			 int (_tv.tv_usec) );
    if (WFS_DEBUG_LEVEL >= 2) {
      if (_augmentor != NULL ) {
	_cur_size += _augmentor->addToOutput(_logfile);
      }
    }
    if (WFS_DEBUG_LEVEL >= 4) {
      _cur_size += fprintf(_logfile, "[[%d]] ", (int) pthread_self() );
    }
    _cur_size += fprintf(_logfile,"[%s] ",db_levels[abs(level)] );
  }
  va_start(args,fmt);
  _cur_size += vfprintf(_logfile,fmt,args);
  va_end(args);
  fflush(_logfile);
  if(!_use_stderr && (_cur_size > _max_size_b)){
    //need to roll
    do_log_rollover();
  }
  pthread_cleanup_pop(0);
  assert(pthread_mutex_unlock(&_ro_lock) == 0);
}

static void *
doListen( void *log ) {
  ((WFSLog *) log)->listen();
}

void
WFSLog::redirectAll() {
  if(_use_stderr) {
    return;
  }

  // Make a socket pair that can be used by non WFS libraries to write
  // to wfs_log
  int sv[2];
  int r = socketpair(AF_UNIX, SOCK_STREAM, 0, sv);
  if( r < 0 ) {
    wfs_log(WFS_DBG_1, "Couldn't open socket pair for redirecting.\n");
    return;
  }

  _local_socket = fdopen( sv[1], "r" );
  assert( pthread_create( &_listener, NULL, &doListen, (void *) this ) == 0 );
  
  _oldstdout = dup(1);
  close(1);
  dup(sv[0]);
  _oldstderr = dup(2);
  close(2);
  dup(sv[0]);
  _other_socket = sv[0];
}

void
WFSLog::listen() {
  
  char *buf = NULL;
  size_t len;
  while(getline(&buf, &len, _local_socket) >= 0) {
    pthread_cleanup_push( &free, (void *) buf );
    wfslogx(WFS_DBGLEVELS, buf);
    pthread_cleanup_pop(0);
  }
  if(buf) {
    free(buf);
  }
}
