/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef __WFS_LOG_H__
#define __WFS_LOG_H__ 1

#include <string>
#include <stdarg.h>

// TODO: Map to semantic levels
enum dbcode {
  WFS_DBG_OFF = 0,
  WFS_DBG_1 = 1, // Critical
  WFS_DBG_2 = 2, // Error
  WFS_DBG_3 = 3, // Info
  WFS_DBG_4 = 4, // Debugging
  WFS_DBGLEVELS,
  WFS_DBG_1_NOHEADER = -WFS_DBG_1, // Critical
  WFS_DBG_2_NOHEADER = -WFS_DBG_2, // Error
  WFS_DBG_3_NOHEADER = -WFS_DBG_3, // Info
  WFS_DBG_4_NOHEADER = -WFS_DBG_4 // Debugging
};

class WFSLogAugmentor {
  public:
    virtual ~WFSLogAugmentor() {};
    /*
     * Has to return the number of characters written or -1 on failure
     */
    virtual int addToOutput(FILE *fp) = 0;
};

/*
 * This class is a non-enforced singleton. It should not be instantiated
 * by users. Instead, it is created/destroyed by WFSLogInit when no other
 * translation units refer to it.
 */
class WFSLog {
  public:
    WFSLog(std::string base_name, std::string log_dir, int max_size, int max_files);
    ~WFSLog();
    void wfslogx(int level,const char* fmt,...);
    void addAugmentor(WFSLogAugmentor * aug) { _augmentor = aug; };
    void redirectAll();
    void listen();

  private:
    void set_logfile(FILE * fp){_logfile = fp;};
    bool do_log_rollover();
    int _max_size_b; // Max log file size converted to bytes
    int _max_files;
    int _cur_size;
    bool _use_stderr;
    int _oldstdout;
    int _oldstderr;
    FILE * _logfile;
    FILE * _local_socket;
    int _other_socket;
    pthread_t _listener;
    std::string _base_path;
    /*
     * Used to add application hooks for debug levels.
     * For BC we only currently support one hook for debug levels
     * over 2. A full feature would allow hooking a class at every
     * level.
     */
    WFSLogAugmentor * _augmentor;

    mutable pthread_mutex_t _ro_lock;
};

/*
 * This class implements the standard trick for making sure that when
 * no other translation units refer to WFSLog, a cleanup is performed
 * (in this case, closes log files). This class should be instantiated
 * as by users of the logging API. This way, its destructor will be called
 * when the module is unloaded (or main exits).
 */
class WFSLogInit
{
  public:
    WFSLogInit(std::string base_name = "wfs.log",
               std::string log_dir =".",
               int max_size = 20,
               int max_files = 10,
               WFSLogAugmentor * aug = NULL);
    ~WFSLogInit();

  private:
    static unsigned int _refcount;
};

extern WFSLog * _wfslog;
extern int WFS_DEBUG_LEVEL;

inline bool wfs_dbg_block(int level)
{
  return (WFS_DEBUG_LEVEL >= level);
}

#define WFS_DBMIN_LEVEL 0 // Minimum supported debug level.
#define WFS_DBMAX_LEVEL WFS_DBGLEVELS // Maximum supported debug level.


#define wfs_log(level,...)                                    \
  do {                                                        \
    if(WFS_DEBUG_LEVEL < abs(level))			      \
    {;}                                                       \
    else {                                                    \
      if(_wfslog) { _wfslog->wfslogx((level),__VA_ARGS__); }  \
    }                                                         \
  } while(0)

void wfs_set_debug(int level);

#endif // __WFS_LOG_H__
