/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#ifndef __WFS_MAINT_DIRDISK_H__ 
#define __WFS_MAINT_DIRDISK_H__ 1

#include <map>
#include <vector>
#include <wfs_client_manager.h>
#include <wfs_storage.h>
#include <wfs_perfile_locker.h>

#include "wfs_maint.h"
#include "wfs_maint_protocol.h"

class wfs_maint_dirdisk : public wfs_maint {

 protected:

  struct repair_data {
    wfs_file_id fid;
    wfs_node node;
    bool pull;
    int32_t explicit_version; // Grab this version of a versioned file
    repair_data *next;
    bool front;
    std::string append_data;
    pthread_mutex_t *sync_lock;
    pthread_cond_t *sync_cond;
    int *sync_left;
    std::vector<wfs_maint_protocol::status> *statuses;
  };

  std::string _server_dir;
  wfs_perfile_locker *_locker;
  node_manager *_nodes;
  wfs_client_manager *_clients;
  pthread_cond_t _repair_cond;
  pthread_mutex_t _repair_lock;
  repair_data *_repair_queue;
  repair_data *_repair_queue_tail;
  repair_data *_repair_queue_frontend;
  bool _maint_started;
  bool _fh_needs_cues;
  std::map<wfs_file_id, std::map<wfs_node_id, std::map<int32_t, repair_data *> > > _repair_map;

  int MAINT_LOOP_TIMER;

  wfs_storage *_storage;

  virtual void get_replicas( wfs_file_handle fh, std::vector<wfs_node> &reps );
  void maint_loop();
  void do_repair( repair_data *d, bool exit, rpcc::TO to = rpcc::to_inf );
  bool do_metadata_exchange( wfs_node node );
  void build_metadata_list( wfs_node_id node_id,
			    std::vector<wfs_maint_protocol::metadata> &md );
  void compare_metadata( wfs_node node, 
			 std::vector<wfs_maint_protocol::metadata> my_md,
			 std::vector<wfs_maint_protocol::metadata> node_md );

  int read_file_data( wfs_file_id fid, int32_t vernum,
		      unsigned long long offset, int size, 
		      std::string &data, std::string &perm_cues, 
		      wfs_version &version, 
		      std::map<std::string,wfs_permission> &permissions,
		      unsigned long &atime,
		      unsigned long &mtime,
		      int &num_links );
  int write_file_data( wfs_file_id fid,
		       unsigned long long offset, 
		       std::string data, std::string perm_cues, 
		       wfs_version version, 
		       const std::map<std::string,wfs_permission> &permissions, 
		       unsigned long atime,
		       unsigned long mtime,
		       int num_links, bool append = false );
  int merge_dirs( wfs_file_id fid, std::string remote_dir );

  virtual wfs_node node_from_id( wfs_node_id id );

 public:

  wfs_maint_dirdisk( wfs_node me, rpcs *server, std::string dir, 
		     wfs_perfile_locker *locker, wfs_storage *storage,
		     wfs_client_manager *clients,
		     int maint_timer = 10 );
  ~wfs_maint_dirdisk();

  void repair_loop();
  void start_maint( node_manager *nodes, bool do_maint = true );
  int exchange_metadata( wfs_node_id node_id,
			 std::vector<wfs_maint_protocol::metadata> in_md,
			 std::vector<wfs_maint_protocol::metadata> &out_md );
  void add_to_repair_queue( wfs_file_id fid, wfs_node node, bool pull,
			    bool add_to_front,
			    int32_t explicit_version = 0);
  int pull_file_data( wfs_file_id fid, int32_t vernum,
		      unsigned long long offset, int32_t size,
		      wfs_maint_protocol::file_data &file_data );
  int push_file_data( wfs_file_id fid,
		      unsigned long long offset,
		      wfs_maint_protocol::file_data file_data, 
		      uint16_t append, int &dummy );
  void immediate_replica_enqueue( wfs_file_handle fh );
  wfs_maint_protocol::status immediate_replica_sync( wfs_file_handle fh, 
						     wfs_file_handle other_fh,
						     std::string append_data = "" );

};

#endif // __WFS_MAINT_DIRDISK_H__ 
