/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include "chan.h"

#include <arpa/inet.h>
#include <malloc.h>

#include "method_thread.h"
#include "logging/wfs_log.h"

const unsigned int CONN_LIMIT = 1000;

cchan::cchan(chan_hub *_hub,
	     sockaddr_in _dst)
  : th (0), dst (_dst), ch (NULL), hub(_hub), lossy_percent(0)
{
  assert(pthread_mutex_init(&connect_m, 0) == 0);
  assert(pthread_mutex_init(&ch_m, 0) == 0);
  assert(pthread_cond_init(&ch_c, 0) == 0);
  assert(pthread_mutex_init(&tcpdelete_m, 0) == 0);
  assert(pthread_mutex_init(&setup_m, 0) == 0);

  deleted = new bool();
  *deleted = false;

  char *loss_env = getenv("RPC_LOSSY");
  if(loss_env != NULL){
    lossy_percent = atoi(loss_env);
  }
}

cchan::~cchan()
{
  wfs_log(WFS_DBG_2, "~cchan start\n");

  assert (pthread_mutex_lock(&setup_m) == 0);
  for( unsigned int i = 0; i < setup_ths.size(); i++ ) {
    wfs_log(WFS_DBG_2, "killing setup thread %x\n", (unsigned) setup_ths[i] );
    assert (pthread_cancel(setup_ths[i]) == 0);
  }
  // wait for connect to be released
  int times = 0;
  while(setup_ths.size() > 0 && times++ < 1000 ) {
    assert (pthread_mutex_unlock(&setup_m) == 0);
    /**
       #ifndef __APPLE__
       pthread_yield();
       #else
       // pthread_yield isn't available on Mac
       pthread_yield_np();
       #endif
    */
    usleep(10000);
    assert (pthread_mutex_lock(&setup_m) == 0);
  }
  *deleted = true;
  assert (pthread_mutex_unlock(&setup_m) == 0);
  assert (pthread_mutex_destroy(&setup_m) == 0);

  // do this after killing the setup threads, in case that created a channel
  if (ch != NULL && th != 0) {
    assert (pthread_cancel(th) == 0);
  }
  if (th) {
    assert (pthread_join(th, NULL) == 0);
    assert (pthread_mutex_lock(&ch_m) == 0);
    th = 0;
    // only if a thread did get started should we potentially wait
    // on it to clean things up
    if (ch) {
      // hasn't been deleted yet (the thread wasn't scheduled
      // long enough to push the cleanup method
      int trials2 = 0;
      while(ch->outstanding() && trials2++ < 1000) {
	usleep(10000);
      }
      delete ch;
    }
    assert(pthread_mutex_unlock(&ch_m) == 0);
  }

  cancel.execute();

  if( times >= 1000 ) {
    // there will be some extra state lying around, sadly; memory leak
    wfs_log(WFS_DBG_2, "~cchan couldn't kill all setup threads, dying anyway\n" );
    // ~xavid: debugging mode shouldn't affect whether we die here
    //assert(0); //-xavid: killing the whole program is excessive?
    return;
  } else {
    delete deleted;
  }

  int res = pthread_mutex_destroy(&connect_m);

  wfs_log(WFS_DBG_2, "~cchan: pthread_mutex_destroy(&connect_m) got a %d: %s\n",
	    res, strerror(res));

  // -xavid: this will fail sometimes, because the connect_m can still be held
  //         if we're deleted in wfs_client_manager due to a failed bind
  //         TODO: fix this more better
  //assert(res == 0);
  assert(pthread_mutex_destroy(&ch_m) == 0);
  assert(pthread_cond_destroy(&ch_c) == 0);
  assert(pthread_mutex_destroy(&tcpdelete_m) == 0);
  wfs_log(WFS_DBG_2, "~cchan done\n");
}

void
cchan::cleanup_setup_thread(pthread_mutex_t *setup_m,
			    std::vector<pthread_t> *setup_ths, pthread_t pth,
			    bool *deleted) {
  if( *deleted ) {
    delete deleted;
    return;  }
  if(pthread_mutex_lock(setup_m) != 0) {
    return; // already been destructed
  }
  // erase my thread from the list
  std::vector<pthread_t>::iterator it;
  for( it = setup_ths->begin(); it != setup_ths->end(); it++ ) {
    if( *it == pth ) {
      wfs_log(WFS_DBG_2, "cleanup_setup_thread erasing %x\n", (unsigned) pth );
      setup_ths->erase(it);
      break;
    }
  }
  assert (pthread_mutex_unlock(setup_m) == 0);
}

void
cchan::setup_and_send(const charbuf pdu)
{
  abstract_channel *s;
  int r = false;
  setup_cleanup_frames();

  wfs_log(WFS_DBG_2, "(%x) starting\n", (unsigned) pthread_self());

  // though this seems unimportant, this lock/unlock ensures that
  // the thread can't finish executing or be cancelled until its
  // parent has placed the thread id on the setup list
  // (otherwise it would never get deleted from the list)
  assert (pthread_mutex_lock(&setup_m) == 0);
  assert (pthread_mutex_unlock(&setup_m) == 0);
  cleanup_data *cd = new cleanup_data(); // freed by the cleanup handler
  cd->setup_m = &setup_m;
  cd->setup_ths = &setup_ths;
  cd->pth = pthread_self();
  cd->deleted = deleted;

  int oldstate;
  assert(pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &oldstate) == 0);
#ifdef WFS_WITH_LIBSSH
  setup_cleanup_handler( &cleanup_frames, frames );
#endif
  setup_cleanup_handler( &cchan::cleanup_setup_thread, cd );
  setup_cleanup_handler( &cb_buf_cleanup, (void *) &pdu  );
  setup_cleanup_handler( &cb_ref_cleanup, (void *) &pdu  );

  wfs_log(WFS_DBG_2, "(%x) setup_and_send called on %s:%d\n",
	    (unsigned) pthread_self(),
	    inet_ntoa(dst.sin_addr), (int)ntohs(dst.sin_port));

  assert(pthread_mutex_lock(&connect_m) == 0);
  cleanup_lock_del *cld = new cleanup_lock_del();
  cld->lock = &connect_m;
  cld->deleted = deleted;
  setup_cleanup_handler( &cchan::cleanup_lock_delcheck, cld );

  assert(pthread_mutex_lock(&ch_m) == 0);
  if (ch != 0)  {
    assert(pthread_mutex_unlock(&ch_m) == 0);
    r = true;
    goto exit;
  }
  assert(pthread_mutex_unlock(&ch_m) == 0);

  if (th != 0) {
    // channel was killed, but we haven't joined with the tcp_loop thread yet
    assert (pthread_join(th, NULL) == 0);
    th = 0;
  }

  s = hub->channel(&cancel, &tcpdelete_m);
  if (s == NULL || *(cld->deleted)) {
    wfs_log(WFS_DBG_2, "chan creation problem\n");
    goto exit;
  }

  int res;
  setup_cleanup_handler( &cchan::cleanup_close_channel, s );
#ifdef WFS_WITH_LIBSSH
  res = s->connect(dst, frames);
#else
  res = s->connect(dst);
#endif
  pop_cleanup_handler(0); //cleanup_close_channel
  if (res < 0 || *(cld->deleted)) {
    wfs_log(WFS_DBG_2, "chan connect problem to %s:%d\n",
	      inet_ntoa(dst.sin_addr), (int)ntohs(dst.sin_port));
    delete s;
    s = NULL;
    goto exit;
  }
  wfs_log(WFS_DBG_2, "(%p) Creating new channel for socket %d\n", this,
		     s->channo() );

  pthread_cleanup_push( &cchan::cleanup_close_channel, (void *)s);
  assert(pthread_mutex_lock(&ch_m) == 0);
  pthread_cleanup_pop(0); //cleanup_close_channel

  pthread_cleanup_push( &cchan::cleanup_lock, (void *) &ch_m );
  ch = s;
  th = method_thread(this, false, &cchan::tcp_loop);
  pthread_cleanup_pop(0); // cleanup_lock
  assert(pthread_mutex_unlock(&ch_m) == 0);
  r = true;

 exit:

  pop_cleanup_handler(1); // cleanup_lock connect_m (also deletes cld)

  wfs_log(WFS_DBG_2, "(%x) setup_and_send setup finished on %s:%d\n",
	    (unsigned) pthread_self(),
	    inet_ntoa(dst.sin_addr), (int)ntohs(dst.sin_port));

  pop_cleanup_handler(0); // ref_cleanup
  pop_cleanup_handler(0); // buf_cleanup

  bool sent = false;
  if(r) {
    assert(pthread_mutex_lock(&ch_m) == 0);
    pthread_cleanup_push( &cchan::cleanup_lock, (void *) &ch_m );
    if (ch != NULL) {
      ch->send(pdu);
      sent = true;
    }
    pthread_cleanup_pop(0); // cleanup_lock
    assert(pthread_mutex_unlock(&ch_m) == 0);
  }

  if(!sent) {
    assert(pthread_mutex_lock(&tcpdelete_m) == 0);
    pdu.decref();
    assert(pthread_mutex_unlock(&tcpdelete_m) == 0);
    if (pdu.free) {
      free(pdu.buf);
    }
  }

  // important to disable so that the thread does get deleted from
  // the list without cancellation
  assert(pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &oldstate) == 0);
  pop_cleanup_handler(1); // cleanup_setup_thread
#ifdef WFS_WITH_LIBSSH
  pop_cleanup_handler(1); // cleanup_frames
#endif

  pthread_exit(0);

}

bool cchan::dead()
{
  bool x;
  assert(pthread_mutex_lock(&ch_m) == 0);
  pthread_cleanup_push( &cchan::cleanup_lock, (void *) &ch_m );
  x = (ch == NULL) || ch->dead();
  pthread_cleanup_pop(0); // cleanup_lock
  assert(pthread_mutex_unlock(&ch_m) == 0);
  return x;
}

void cchan::done()

{
  int oldstate;
  assert(pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &oldstate) == 0);
  assert(pthread_mutex_lock(&ch_m) == 0);
  pthread_cleanup_push( &cchan::cleanup_lock, (void *) &ch_m );
  wfs_log(WFS_DBG_2, "cchan (%p) starting done\n", this );
  if(ch) {
    int trials = 0;
    while(ch->outstanding() && trials++ < 1000) {
      usleep(10000);
    }
    delete ch;
    ch = NULL;
  }
  wfs_log(WFS_DBG_2, "cchan (%p) ending done\n", this );
  pthread_cleanup_pop(0); // cleanup_lock
  assert(pthread_mutex_unlock(&ch_m) == 0);
}

void
cchan::send(const charbuf pdu)
{
  bool is_setup;
  assert(pthread_mutex_lock(&ch_m) == 0);
  // reject if this channel is already dead
  is_setup = (ch != NULL);
  assert(pthread_mutex_unlock(&ch_m) == 0);

  // done in same thread where it was created, so no need for a lock
  pdu.incref();

  if (is_setup) {
    if (lossy_percent) {
      if((random() % 100) < lossy_percent) {
	pdu.decref();
        return;
      }
      if((random() % 100) < 10) {
        sleep(random() % 10); // delay request
      }
      if((random() % 100) < lossy_percent) {
	pdu.incref();  // double send
        ch->send(pdu);
      }
    }
    assert(pthread_mutex_lock(&ch_m) == 0);
    pthread_cleanup_push( &cchan::cleanup_lock, (void *) &ch_m );
    if (ch != NULL) {
      ch->send(pdu);
    } else {
      pdu.decref();
    }
    pthread_cleanup_pop(0); // cleanup_lock
    assert(pthread_mutex_unlock(&ch_m) == 0);
  } else {
    assert (pthread_mutex_lock(&setup_m) == 0);
    pthread_cleanup_push( &cchan::cleanup_lock, (void *) &setup_m );
    pthread_t th = method_thread(this, true, &cchan::setup_and_send, pdu);
    setup_ths.push_back(th);
    // if we have more than one thread attempting a setup, let's kill the
    // earlier one to clean up the connect lock -- the rpc layer thinks we've
    // waited long enough
    if (setup_ths.size() > 1) {
	    wfs_log(WFS_DBG_2, "killing a thread (%x) due to timeout!\n",
		    (unsigned)setup_ths[setup_ths.size()-2]);
      assert (pthread_cancel(setup_ths[setup_ths.size()-2]) == 0);
    }
    pthread_cleanup_pop(0); // cleanup_lock
    assert (pthread_mutex_unlock(&setup_m) == 0);
  }
}

// copy PDUs from a tcpchan to inq
void
cchan::tcp_loop()
{

  setup_cleanup_frames();
  pthread_cleanup_push(&cchan::cleanup_tcp_loop, (void *) this);

  int oldstate;
  assert(pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &oldstate) == 0);
  //assert(pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &oldstate) == 0);

  while(1) {
    charbuf cb;
    wfs_log(WFS_DBG_2, "(%x) cchan::tcploop receiving\n",
			 (unsigned) pthread_self() );
#ifdef WFS_WITH_LIBSSH
    int s = ch->recv(cb, frames);
#else
    int s = ch->recv(cb);
#endif
    if (ch->dead()) {
      break;
    } else if( s > 0 ) {
      inq.enq(cb);
    }
  }
  pthread_cleanup_pop(1);
}

void
cchan::recv(charbuf &cb)
{
  cb = inq.deq();
  return;
}

schan::schan(chan_hub *_hub, int port)
  : waiting (false), hub(_hub), lossy_percent(0)
{
  assert(pthread_mutex_init(&tcpchans_m, 0) == 0);
  assert(pthread_cond_init(&tcpchans_c, 0) == 0);
  assert(pthread_cond_init(&outstanding_c, 0) == 0);

  char *loss_env = getenv("RPC_LOSSY");
  if(loss_env != NULL){
    lossy_percent = atoi(loss_env);
  }

  struct sockaddr_in sin;
  bzero(&sin, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_port = port;

  tcp = hub->bind(sin);
  if (tcp == NULL){
    wfs_log(WFS_DBG_2,"schan hub.bind: failed to bind to port %d",port);
    perror("");
    exit(1);
  }

  th_tcp_loop = method_thread(this, false, &schan::tcp_loop);

}

schan::~schan()
{
  wfs_log(WFS_DBG_2, "~schan start\n");
  tcp->close();
  assert(pthread_mutex_lock(&tcpchans_m) == 0);
  std::map<int,channel>::iterator iter;
  for (iter = tcpchans.begin(); iter != tcpchans.end(); iter++) {
    abstract_channel *ch = iter->second.ch;
    pthread_t th = iter->second.th;
    if (ch) ch->die();
    pthread_cancel(th);
  }
  assert(pthread_mutex_unlock(&tcpchans_m) == 0);
  assert(pthread_cancel(th_tcp_loop)==0);
  pthread_join(th_tcp_loop, NULL);

  assert(pthread_mutex_lock(&tcpchans_m) == 0);
  while (tcpchans.size() > 0) {
    waiting = true;
    wfs_log(WFS_DBG_2, "~schan waiting for tcpchan disappear %u\n",
	     (unsigned)tcpchans.size());
    assert(pthread_cond_wait(&tcpchans_c, &tcpchans_m) == 0);
  }
  wfs_log(WFS_DBG_2, "~schan after wait\n");
  assert(pthread_mutex_unlock(&tcpchans_m) == 0);

  cancel.execute();

  assert(pthread_mutex_destroy(&tcpchans_m) == 0);
  assert(pthread_cond_destroy(&tcpchans_c) == 0);
  assert(pthread_cond_destroy(&outstanding_c) == 0);
  wfs_log(WFS_DBG_2, "~schan done\n");
}

// wait for each new tcp connection, turn it into a tcpchan,
// start a thread to read from that tcpchan.
void
schan::tcp_loop()
{
  int oldstate;
  assert(pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &oldstate) == 0);
  //assert(pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &oldstate) == 0);

  wfs_log(WFS_DBG_2, "starting tcp_loop from thread %x...\n",
      (unsigned)pthread_self());

  while(1) {
    // TODO: limit the number of open connections
    pthread_mutex_t *delete_lock = new pthread_mutex_t();
    assert(pthread_mutex_init(delete_lock, NULL) == 0);
    abstract_channel *s1 = tcp->accept(&cancel, delete_lock);
    if(s1 == NULL){
      perror("schan accept");
      assert(pthread_mutex_destroy(delete_lock) == 0);
      delete delete_lock;
      continue;
    }
    wfs_log(WFS_DBG_2, "schan accepted socket %d\n", s1->channo());

    method_thread(this, true, &schan::tcp_loop1, s1, delete_lock);
  }
}

void schan::done(pthread_t myth)
{
  abstract_channel *ch = NULL;

  assert (pthread_mutex_lock(&tcpchans_m) == 0);
  std::map<int,channel>::iterator iter;
  for (iter = tcpchans.begin(); iter != tcpchans.end(); iter++) {
    pthread_t th = iter->second.th;
    if (th == myth) {
      ch = iter->second.ch;
      break;
    }
  }

  assert (ch);
  wfs_log(WFS_DBG_2, "done with socket %d\n", iter->first);
  lru_sockets.remove(iter->first);
  cleanup *obj = NULL;
  int channo = iter->first;

  // we can't leave anything queued for this channel (so that we can
  // safely reuse the channo)
  while (iter->second.outstanding > 0) {
    // make sure the outstanding packet won't try to double cancel the thread,
    // and that no new packets will be processed
    iter->second.th = 0;
    iter->second.die = true;
    pthread_cond_wait(&outstanding_c, &tcpchans_m);
  }

  if(iter->second.clnup != NULL) {
    obj = iter->second.clnup;
  }

  tcpchans.erase(iter);
  if (waiting)
    assert(pthread_cond_signal(&tcpchans_c) == 0);
  assert(pthread_mutex_unlock(&tcpchans_m) == 0);

  int trials = 0;
  while (ch->outstanding() && trials < 1000) {
    /**
#ifndef __APPLE__
    pthread_yield();
#else
    // pthread_yield isn't available on Mac
    pthread_yield_np();
#endif
    */
    usleep(10000);
    trials++;
  }
  if(trials >= 1000) {
    wfs_log(WFS_DBG_2, "Gave up on outstanding sends for socket %d\n", channo );
  }

  // delete before cleanup, otherwise the cleanup could delete some replies
  // already on the tcpchan queue
  delete ch;

  // avoid holding tcpchans_m during cleanup
  if (obj != NULL) {
    obj->do_cleanup(channo);
  }

  assert(pthread_mutex_lock(&tcpchans_m) == 0);
  assert(pthread_mutex_destroy(delete_locks[channo])==0);
  delete delete_locks[channo];
  delete_locks[channo] = NULL;
  delete_locks.erase(channo);
  assert(pthread_mutex_unlock(&tcpchans_m) == 0);

  wfs_log(WFS_DBG_2, "finished done with socket %d\n", channo);

}

void
schan::lock_for_delete(int channo) {
  assert(pthread_mutex_lock(&tcpchans_m) == 0);
  assert(delete_locks.find(channo) != delete_locks.end());
  pthread_mutex_t *m = delete_locks[channo];
  assert(pthread_mutex_unlock(&tcpchans_m) == 0);
  assert(pthread_mutex_lock(m) == 0);
}

void
schan::unlock_for_delete(int channo) {
  assert(pthread_mutex_lock(&tcpchans_m) == 0);
  assert(delete_locks.find(channo) != delete_locks.end());
  pthread_mutex_t *m = delete_locks[channo];
  assert(pthread_mutex_unlock(&tcpchans_m) == 0);
  assert(pthread_mutex_unlock(m) == 0);
}

bool schan::verify_user(const char *username, int channo) {
  bool ret;
  assert(pthread_mutex_lock(&tcpchans_m) == 0);
  pthread_cleanup_push( &schan::cleanup_lock, (void *) &tcpchans_m );
  ret = tcpchans[channo].ch->verify_user(username);
  pthread_cleanup_pop(0);
  assert(pthread_mutex_unlock(&tcpchans_m) == 0);
  return ret;
}

// just accept()ed a connection to an schan.
// wait for input on a single tcpchan, add to inq.
// blocks inside tcpchan::recv().
void
schan::tcp_loop1(abstract_channel *ch, pthread_mutex_t *delete_lock)
{
  pthread_t th = pthread_self();

  // unfortunately the socket could be available even if the old
  // channel attached to that socket hasn't finished cleaning up
  // yet.  wait, so we can reuse that lock slot
  // we do this in a separate thread from accept() in order to avoid
  // blocking future accepts.  This can happen when there are outstanding
  // sends on an a reused file descriptor.
  bool redo = false;
  do {
    assert(pthread_mutex_lock(&tcpchans_m) == 0);
    if (delete_locks.find(ch->channo()) == delete_locks.end()) {
      delete_locks[ch->channo()] = delete_lock;
      redo = false;
    } else {
      redo = true;
    }
    assert(pthread_mutex_unlock(&tcpchans_m) == 0);
    if (redo) {
      usleep(10000);
    }
  } while(redo);

  int oldstate;
  assert(pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &oldstate) == 0);
  //assert(pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &oldstate) == 0);

  //rpc server's tcpchan free the buffer size after sending

  int channo = ch->channo();
  wfs_log(WFS_DBG_2, "Creating new socket %d (%s)\n", channo,
	   ch->getpeername());

  assert(pthread_mutex_lock(&tcpchans_m) == 0);
  assert(tcpchans.count(channo) == 0);
  tcpchans[channo].ch = ch;
  tcpchans[channo].th = th;
  tcpchans[channo].clnup = NULL;
  tcpchans[channo].outstanding = 0;
  tcpchans[channo].die = false;
  lru_sockets.push_front(channo);

  // if we have too many connections, kill the oldest one
  if (lru_sockets.size() > CONN_LIMIT) {
    int remove_s = lru_sockets.back();
    tcpchans[remove_s].die = true;
    // need to remove it here, even though it's also done in schan::done
    lru_sockets.pop_back();
    if (tcpchans[remove_s].outstanding == 0 ) {
      pthread_t th = tcpchans[remove_s].th;
      wfs_log(WFS_DBG_2, "Connection limit hit; killing socket %d (while creating %d)\n",
          remove_s, channo );
      assert(pthread_cancel(th) == 0);
    }
  }

  assert(pthread_mutex_unlock(&tcpchans_m) == 0);

  setup_cleanup_frames();
#ifdef WFS_WITH_LIBSSH
  setup_cleanup_handler( &cleanup_frames, frames );
#endif

  pthread_cleanup_push(&schan::cleanup_tcp_loop1, (void *) this);

  if (ch->start() != 0) {
    wfs_log(WFS_DBG_2, "Start failed on socket %d, cleaning up\n", channo );
    goto exit;
  }

  while(!ch->dead()){
    if(lossy_percent > 0 && (random() % 100) < lossy_percent)
      ch->lose();
    charbuf pdu;
#ifdef WFS_WITH_LIBSSH
    if (ch->recv(pdu, frames) > 0) {
#else
    if (ch->recv(pdu) > 0) {
#endif
      assert(pthread_mutex_lock(&tcpchans_m) == 0);
      pthread_cleanup_push( &schan::cleanup_lock, (void *) &tcpchans_m );
      if (!tcpchans[channo].die) {
	lru_sockets.remove(channo);
	lru_sockets.push_front(channo);
	tcpchans[channo].outstanding++;
	inq.enq(inbuf(pdu, channo));
      }
      pthread_cleanup_pop(0);
      assert(pthread_mutex_unlock(&tcpchans_m) == 0);
    } else if (ch->dead()) {
      assert(pthread_mutex_lock(&tcpchans_m) == 0);
      pthread_cleanup_push( &schan::cleanup_lock, (void *) &tcpchans_m );
      lru_sockets.remove(channo); // make sure it's not double-canceled later
      pthread_cleanup_pop(0);
      assert(pthread_mutex_unlock(&tcpchans_m) == 0);
      break;
    }
  }

 exit:
  pthread_cleanup_pop(1);
#ifdef WFS_WITH_LIBSSH
  pop_cleanup_handler(1); // cleanup frames
#endif
}

// chan should be passed again to send(), to ensure that
// RPC reply goes back to where the request came from.
void
schan::recv(charbuf &pdu, int &channo)
{
  inbuf b = inq.deq();
  pdu = b.s;
  channo = b.channo;
}

void
schan::send(const charbuf pdu, int channo)
{
  abstract_channel *ch = 0;
  assert(pthread_mutex_lock(&tcpchans_m) == 0);
  if (tcpchans.count(channo) > 0)
    ch = tcpchans[channo].ch;
  assert(pthread_mutex_unlock(&tcpchans_m) == 0);

  if (!ch) {
    return;
  }

  pdu.incref();

  // only send the packet if it's not empty (an empty packet indicates
  // that there will not be a real response to an outstanding RPC)
  assert(pdu.sz >= 0);
  if (pdu.sz > 0) {
    if (lossy_percent) {
      if((random() % 100) < lossy_percent) {
	return;
      }
      if((random() % 100) < 10) {
	sleep(random() % 10); // delay request
      }
      if((random() % 100) < lossy_percent) {
	charbuf pdudup = pdu.dup();
	pdudup.free = true;
	pdu.incref();
	ch->send(pdudup);
      }
    }
    ch->send(pdu);
  }

  assert(pthread_mutex_lock(&tcpchans_m) == 0);
  if (tcpchans.count(channo) > 0) {
    tcpchans[channo].outstanding--;
    if (tcpchans[channo].outstanding == 0 && tcpchans[channo].die &&
	tcpchans[channo].th != 0) {
      pthread_t th = tcpchans[channo].th;
      wfs_log(WFS_DBG_2, "Connection limit hit; killing socket %d (send)\n",
          channo );
      assert(pthread_cancel(th) == 0);
    }
    pthread_cond_broadcast(&outstanding_c);
  }
  assert(pthread_mutex_unlock(&tcpchans_m) == 0);

}

bool
schan::set_cleanup(int channo, cleanup* obj) {
  bool used = true;
  assert(pthread_mutex_lock(&tcpchans_m) == 0);
  wfs_log(WFS_DBG_2, "(chan) Setting cleanup object for channo %d\n", channo );
  if (tcpchans.find(channo) != tcpchans.end()) {
    assert(tcpchans[channo].clnup == NULL);
    tcpchans[channo].clnup = obj;
  } else {
    used = false;
  }
  assert(pthread_mutex_unlock(&tcpchans_m) == 0);
  return used;
}

