/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


// server worker thread pool
// TCP transport (probably needed for good wide-area behavior)
// libarpc lets you send and serve on the same port, do we need that?

#include "rpc.h"
#include "method_thread.h"
#ifdef VIVALDI
#include "vivaldi.h"
#endif
#include <arpa/inet.h>
#include <sys/select.h>
#include <time.h>
#include <netdb.h>

#include "chan.h"
#include "tcpchan.h"
#include "sshchan.h"
#include "logging/wfs_log.h"

const int PER_CHANNEL_OUTSTANDING = 10;

rpcc::rpcc(sockaddr_in _dst)
  : dst(_dst), xid(1), svr_nonce(0), bind_done(false),
    _vivaldi(NULL), destroy_wait (false)
{
  //fprintf(stderr,"temp rpcc: constructed with %s\n",inet_ntoa(dst.sin_addr));
  static chan_hub *hub = NULL;
  if (hub == NULL) {
    if (!rpc_is_secure())
      hub = new tcphub();
    else {
#ifdef WFS_WITH_LIBSSH
      hub = new sshhub();
#else
      assert (false);
#endif
    }
  }
  chan = new cchan(hub, dst);

  assert(pthread_mutex_init(&m, 0) == 0);
  assert(pthread_mutex_init(&_sender_lock, 0) == 0);
  assert(pthread_mutex_init(&_timeout_lock, 0) == 0);
  assert(pthread_cond_init(&_timeout_cond, 0) == 0);
  assert(pthread_cond_init(&destroy_wait_c, 0) == 0);
  _next_timeout.tv_sec = 0;

  if((th_clock_loop = method_thread(this, false, &rpcc::clock_loop)) == 0){
    perror("rpcc::rpcc pthread_create");
    exit(1);
  }
  if((th_chan_loop = method_thread(this, false, &rpcc::chan_loop)) == 0){
    perror("rpcc::rpcc pthread_create");
    exit(1);
  }

  // assume random seeding done outside of the RPC library
  clt_nonce = random();

  xid_rep_window.push_back(0);

  wfs_log(WFS_DBG_2, "rpcc: my nonce %d\n", clt_nonce);

}

rpcc::~rpcc()
{

  wfs_log(WFS_DBG_2, "~rpcc\n");

  assert(pthread_cancel(th_clock_loop) == 0);
  assert(pthread_cancel(th_chan_loop) == 0);

  assert(pthread_join(th_clock_loop, NULL) == 0);
  assert(pthread_join(th_chan_loop, NULL) == 0);

  calls.clear();
  xid_rep_window.clear();

  assert(pthread_mutex_destroy(&m) == 0);
  assert(pthread_mutex_destroy(&_sender_lock) == 0);
  int r;
  if ((r = pthread_mutex_destroy(&_timeout_lock)) != 0) {
    wfs_log(WFS_DBG_OFF, "timeout_lock: error? %d busy? %d\n", r==EINVAL?1:0, r==EBUSY?1:0);
  }
  if ((pthread_cond_destroy(&_timeout_cond)) != 0) {
    wfs_log(WFS_DBG_OFF, "timeout_cond: error? %d busy? %d\n", r==EINVAL?1:0, r==EBUSY?1:0);
  }
  assert(pthread_cond_destroy(&destroy_wait_c) == 0);
  wfs_log(WFS_DBG_2, "~rpcc done\n");
  delete chan;
}

void
rpcc::setlossy(bool x)
{
  if (x)
    chan->setlossy();
  else
    chan->setlossy(0);
}

void
rpcc::cancel(void)
{
  assert(pthread_mutex_lock(&m) == 0);
  std::map<int,caller*>::iterator iter;
  for(iter = calls.begin(); iter != calls.end(); iter++){
    caller *ca = iter->second;
    wfs_log(WFS_DBG_2, "cancel: force caller to fail\n");
    assert(pthread_mutex_lock(&ca->m) == 0);
    ca->done = true;
    ca->intret = rpc_const::cancel_failure;
    assert(pthread_cond_signal(&ca->c) == 0);
    assert(pthread_mutex_unlock(&ca->m) == 0);
  }

  while (calls.size () > 0) {
    destroy_wait = true;
    assert(pthread_cond_wait(&destroy_wait_c,&m) == 0);
  }

  assert(pthread_mutex_unlock(&m) == 0);
}

// bind this rpc to a rpcs instance; should be done exactly once.
// necessary to ensure at-most-once RPC semantics.
int
rpcc::bind(TO to)
{
  int r;
  int ret = call(rpc_const::bind, 0, r, to);
  if (ret == 0) {
    assert(pthread_mutex_lock(&m) == 0);
    svr_nonce = r;
    bind_done = true;
    assert(pthread_mutex_unlock(&m) == 0);
  }
  return ret;
};

rpcc::caller::caller(int xxid, unmarshall *xun, uint32_t ip, uint16_t port, double when)
  : xid(xxid), un(xun), done(false), senttime(when), other_ip(ip), other_port(port)
{
  assert(pthread_cond_init(&c, 0) == 0);
  assert(pthread_mutex_init(&m, 0) == 0);
}

rpcc::caller::~caller()
{
  assert(pthread_mutex_destroy(&m) == 0);
  assert(pthread_cond_destroy(&c) == 0);
}

const rpcc::TO rpcc::to_inf = { -1, false };

/* Subtract the `struct timeval' values X and Y,
   storing the result in RESULT.
   Return 1 if the difference is negative, otherwise 0.  */
// taken from http://www.delorie.com/gnu/docs/glibc/libc_428.html
int
timeval_subtract (struct timeval *result, struct timeval *x, struct timeval *y)
{
  struct timeval y2 = *y;

  /* Perform the carry for the later subtraction by updating y. */
  if (x->tv_usec < y2.tv_usec) {
    int nsec = (y2.tv_usec - x->tv_usec) / 1000000 + 1;
    y2.tv_usec -= 1000000 * nsec;
    y2.tv_sec += nsec;
  }
  if (x->tv_usec - y2.tv_usec > 1000000) {
    int nsec = (x->tv_usec - y2.tv_usec) / 1000000;
    y2.tv_usec += 1000000 * nsec;
    y2.tv_sec -= nsec;
  }

  /* Compute the time remaining to wait.
     tv_usec is certainly positive. */
  result->tv_sec = x->tv_sec - y2.tv_sec;
  result->tv_usec = x->tv_usec - y2.tv_usec;

  /* Return 1 if result is negative. */
  return x->tv_sec < y2.tv_sec;
}

void
add_timeout(struct timeval a, int b, struct timespec *result,
	    struct timeval *result_tv) {
  // convert to millisec, add timeout, convert back
  long double msec = (a.tv_sec*1000.0 + a.tv_usec/1000.0)+b*1.0;
  result->tv_sec = (long) (msec/1000);
  result->tv_nsec =
    ((long) ((msec-((long double) result->tv_sec*1000.0))*1000))*1000;
  result_tv->tv_sec = result->tv_sec;
  result_tv->tv_usec = result->tv_nsec/1000;
}

int
rpcc::call1(unsigned int proc, marshall &req, unmarshall &rep,
	    TO to, const char *required_user)
{

  assert(pthread_mutex_lock(&_sender_lock) == 0);
  bool send_locked = true;
  assert(pthread_mutex_lock(&m) == 0);
  if ((proc != rpc_const::bind && !bind_done) ||
      (proc == rpc_const::bind && bind_done)) {
    wfs_log(WFS_DBG_OFF, "rpcc::call1 rpcc has not been bound to server or binding twice\n");
    assert (pthread_mutex_unlock(&m) == 0);
    return rpc_const::bind_failure;
  }

  if (destroy_wait) {
    assert (pthread_mutex_unlock(&m) == 0);
    return rpc_const::cancel_failure;
  }

  unsigned int myxid = xid++;

  caller ca(myxid, &rep, dst.sin_addr.s_addr, dst.sin_port,
#ifdef VIVALDI
            (proc & rpc_const::VIVALDI_MASK)?vivaldi::gettime():-0.1
#else
            -0.1
#endif
            );
  calls[myxid] = &ca;
  assert(pthread_mutex_unlock(&m) == 0);

  req.zero_ind();
  req << clt_nonce << svr_nonce << proc << myxid << xid_rep_window.front();
  assert(req.ind() <= RPC_HEADER_SZ);
  req.reset_ind();

  // ~xavid: get the size first, since marshall::take_buf() changes it
  int cbsize = req.size();
  charbuf cb;
  if (req.del_buf()) {
    cb = charbuf(req.take_buf(), cbsize,
		 required_user);
  } else {
    cb = charbuf(req.cstr(), cbsize,
		 required_user);
  }
  cb.refcount = (int *) malloc(sizeof(int));
  *(cb.refcount) = 1;

  struct timeval first;
  gettimeofday(&first, NULL);
  struct timeval last = first;
  int initial_rto = chan->initial_timeout();

  int rto = initial_rto;
  int next_rto = rto;

  struct timespec deadline;
  struct timeval deadline_tv;
  if(to.to != -1) {
    add_timeout(first, to.to, &deadline, &deadline_tv);
  }

  assert (pthread_mutex_lock(&ca.m) == 0);
  struct timeval now, diff;
  now = last;

  wfs_log(WFS_DBG_2,
	    "rpcc::call %u sending proc %x xid %d %s:%d timeout %d ms rto %d ms\n",
	    clt_nonce, proc, myxid,
	    inet_ntoa(dst.sin_addr),
	    ntohs(dst.sin_port), to.to, rto);

  while(1){
    if(ca.done)
      break;
    if(timeval_subtract(&diff, &now, &last) != 0) {
      // I think this can only occur when NTP resets the system clock
      // during an RPC.  Annoying, let's just call it a timeout.
      wfs_log(WFS_DBG_2, "rpcc::call %u xid %d got a bad "
			 "timeval_subtract, considering it a timeout\n",
			 clt_nonce, myxid );
      break;
    }
    long double diff_msec = diff.tv_sec*1000.0 + diff.tv_usec/1000.0;
    if(diff_msec >= rto || next_rto == initial_rto){
      rto = next_rto;
      if(rto != initial_rto) {
        wfs_log(WFS_DBG_2,
            "rpcc::call %u retransmit proc %x xid %d %s:%d timeout %d ms\n",
            clt_nonce, proc, myxid,
            inet_ntoa(dst.sin_addr),
            ntohs(dst.sin_port), rto);
      }
      chan->send(cb);
      // after the first send, we can release the sender lock
      if( send_locked ) {
	assert(pthread_mutex_unlock(&_sender_lock) == 0);
	send_locked = false;
      }
      gettimeofday(&last, NULL);
      // increase rxmit timer for next time
      next_rto *= 2;
      if(next_rto > 128000) {
	next_rto = 128000;
      }
    }

    // rexmit deadline
    struct timespec my_deadline;
    struct timeval my_deadline_tv;
    add_timeout(last, rto, &my_deadline, &my_deadline_tv);

    // my next deadline is either for rxmit or timeout
    if(to.to != -1 &&
       timeval_subtract(&diff, &deadline_tv, &my_deadline_tv) > 0) {
      my_deadline = deadline;
      my_deadline_tv = deadline_tv;
    }

    assert(pthread_mutex_lock(&_timeout_lock) == 0);
    struct timeval nt_tv;
    nt_tv.tv_sec = _next_timeout.tv_sec;
    nt_tv.tv_usec = _next_timeout.tv_nsec/1000;
    if(_next_timeout.tv_sec == 0 ||
       timeval_subtract(&diff, &my_deadline_tv, &nt_tv) > 0){
      _next_timeout = my_deadline;
      pthread_cond_signal(&_timeout_cond);
    }
    assert(pthread_mutex_unlock(&_timeout_lock) == 0);

    // wait for reply or timeout
    assert(pthread_cond_wait(&ca.c, &ca.m) == 0);

    // user-specified timeout occurred
    gettimeofday(&now, NULL);
    if(!ca.done && to.to != -1){
      if(timeval_subtract(&diff, &deadline_tv, &now) > 0) {
	break;
      }
    }

  }

  if( send_locked ) {
    assert(pthread_mutex_unlock(&_sender_lock) == 0);
    send_locked = false;
  }

  int intret = ca.done ? ca.intret : rpc_const::timeout_failure;
  wfs_log(WFS_DBG_2,
	    "rpcc::call %u done sending proc %x xid %d %s:%d res %d timeout? %d\n",
	    clt_nonce, proc, myxid,
	    inet_ntoa(dst.sin_addr),
	    ntohs(dst.sin_port), intret, intret == rpc_const::timeout_failure);
  pthread_mutex_unlock(&ca.m);

  assert(pthread_mutex_lock(&m) == 0);
  calls.erase(myxid);
  if (destroy_wait) {
    assert(pthread_cond_signal(&destroy_wait_c) == 0);
  }
  // we potentially need to update the xid again here, in case the
  // packet times out before it's even sent by the channel.  nasty.
  // but I don't think there's any harm in potentially doing it twice
  update_xid_rep(myxid);
  assert(pthread_mutex_unlock(&m) == 0);

  // don't want to delete it while the send is happening, that's why
  // we lock . . .
  chan->lock_for_delete();
  wfs_log(WFS_DBG_2, "refcount is %d (%u %d)\n", *(cb.refcount),
		     clt_nonce, myxid );
  cb.decref(true);
  if( req.del_buf() ) {
    free(cb.buf);
  }
  chan->unlock_for_delete();
  return intret;
}

// listen for replies, hand to xid's waiting rpcc::call().
void
rpcc::chan_loop()
{
  th_chan_loop = pthread_self();
  int oldstate;
  assert(pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &oldstate) == 0);

  while(1){
    charbuf cb;
    chan->recv(cb);
#ifdef WFS_RPC_CHECKSUMMING
    wfs_log(WFS_DBG_2, "got reply with size %d and checksum %llu\n",
	      cb.sz, cb.checksum());
#endif
    unmarshall rep(cb.buf, cb.sz, true);
    got_reply(rep);
  }
}

// periodically wakes up all callers so they can think
// about retransmitting.
void
rpcc::clock_loop()
{
  th_clock_loop = pthread_self();
  int oldstate;
  assert(pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &oldstate) == 0);

  while(1){

    // if there are no timeouts, wake until there is one
    assert(pthread_mutex_lock(&_timeout_lock) == 0);
    bool signalled = false;
    pthread_cleanup_push(&rpcc::cleanup_timeout_lock,
         (void *) &_timeout_lock);
    if(_next_timeout.tv_sec == 0) {
      pthread_cond_wait(&_timeout_cond, &_timeout_lock);
      signalled = true;
    } else {
      signalled =
	(pthread_cond_timedwait(&_timeout_cond, &_timeout_lock, &_next_timeout)
	 != ETIMEDOUT);
    }
    pthread_cleanup_pop(0);

    // if we were signalled to be woken up, that means our timeout value
    // was set to something new, so we just need to wait again, and not
    // bother waking everyone up
    if(!signalled)
      _next_timeout.tv_sec = 0;
    assert(pthread_mutex_unlock(&_timeout_lock) == 0);
    if(signalled)
      continue;

    assert(pthread_mutex_lock(&m) == 0);
    std::map<int,caller*>::iterator iter;
    for(iter = calls.begin(); iter != calls.end(); iter++){
      caller *ca = iter->second;
      assert(pthread_mutex_lock(&ca->m) == 0);
      assert(pthread_cond_broadcast(&ca->c) == 0);
      assert(pthread_mutex_unlock(&ca->m) == 0);
    }
    assert(pthread_mutex_unlock(&m) == 0);
  }
}

void
rpcc::got_reply(unmarshall &rep)
{
  int xid, intret;
  rep.zero_ind();
  rep >> xid;
  if(!rep.ok()) {
    return;
  }

  assert(pthread_mutex_lock(&m) == 0);

  if(calls.count(xid) < 1){
    update_xid_rep(xid);
    assert(pthread_mutex_unlock(&m) == 0);
    return;
  }

  caller *ca = calls[xid];
  update_xid_rep(xid);

  rep >> intret;

  assert(pthread_mutex_lock(&ca->m) == 0);

#ifdef VIVALDI
  if (ca->senttime > 0.0) {
	 Coord c;
	 rep >> c;
	 assert(_vivaldi);
	 double now = vivaldi::gettime();
	 _vivaldi->update_other_coords(ca->other_ip, ca->other_port, c);
	 _vivaldi->update_other_lat (ca->other_ip, now - ca->senttime);
	 _vivaldi->update_coords(c, now - ca->senttime);
  }
#endif

  assert(rep.ind() <= RPC_HEADER_SZ);
  rep.reset_ind();

  ca->un->take_in(rep);
  ca->intret = intret;
  ca->done = 1;
  assert(pthread_cond_broadcast(&ca->c) == 0);
  assert(pthread_mutex_unlock(&ca->m) == 0);

  // Wait until down here to unlock the global lock.  This avoids the
  // case where got_reply() gets the caller from the map and locks ca->m,
  // then call1() erases the caller from the map and tries to destroy the
  // caller object before ca->m is unlocked, causing disaster.
  assert(pthread_mutex_unlock(&m) == 0);
}

// assumes thread holds mutex m
void rpcc::update_xid_rep(unsigned int xid)
{
  std::list<unsigned int>::iterator it;

  if (xid <= xid_rep_window.front()) {
    return;
  }

  for (it = xid_rep_window.begin(); it != xid_rep_window.end(); it++) {
    if (*it > xid) {
      xid_rep_window.insert(it, xid);
      goto compress;
    }
  }
  xid_rep_window.push_back(xid);

 compress:
  it = xid_rep_window.begin();
  for (it++; it != xid_rep_window.end(); it++) {
    if (xid_rep_window.front() + 1 == *it)
      xid_rep_window.pop_front();
  }
}


rpcs::rpcs(unsigned int port)
  : lossy(false), lossy_percent(5), _vivaldi(NULL),
    deleting(false), counting(0)
{
  static chan_hub *hub = NULL;
  if (hub == NULL) {
    if (!rpc_is_secure())
      hub = new tcphub();
    else {
#ifdef WFS_WITH_LIBSSH
      hub = new sshhub();
#else
      assert (false);
#endif
    }
  }
  chan = new schan(hub, port);

  assert(pthread_mutex_init(&procs_m, 0) == 0);
  assert(pthread_mutex_init(&reply_window_m, 0) == 0);

  assert(pthread_mutex_init(&delete_m, 0) == 0);
  assert(pthread_cond_init(&delete_c, 0) == 0);

  // assume random seeding happens outside of the RPC library
  nonce = random();

  assert(pthread_mutex_init(&nt_lock, 0) == 0);
  assert(pthread_cond_init(&nt_cond, 0) == 0);

  char *loss_env = getenv("RPC_LOSSY");
  if(loss_env != NULL){
    lossy_percent = atoi(loss_env);
    if(lossy_percent > 0){
      setlossy(true);
    }
  }

  if((th_loop = method_thread(this, false, &rpcs::loop)) == 0){
    perror("rpcs::rpcs pthread_create");
    exit(1);
  }

  reg(rpc_const::bind, this, &rpcs::bind, true);
}

rpcs::~rpcs()
{

  wfs_log(WFS_DBG_2, "~rpcs\n");
  assert (pthread_cancel(th_loop) == 0);
  assert (pthread_mutex_lock(&delete_m) == 0);
  deleting = true;
  while (num_threads.size() > 0) {
    wfs_log(WFS_DBG_2, "~rpcs: wait for handlers\n");
    assert (pthread_cond_wait(&delete_c, &delete_m) == 0);
  }
  assert (pthread_mutex_unlock(&delete_m) == 0);
  pthread_join(th_loop, NULL);

  delete chan;

  free_reply_window();

  // delete all handlers
  assert(pthread_mutex_lock(&procs_m) == 0);
  std::map<int, handler *>::iterator proc_it;
  for( proc_it = procs.begin(); proc_it != procs.end(); proc_it++ ) {
    delete proc_it->second;
  }
  assert(pthread_mutex_unlock(&procs_m) == 0);

  assert(pthread_mutex_destroy(&delete_m) == 0);
  assert(pthread_cond_destroy(&delete_c) == 0);
  assert(pthread_mutex_destroy(&reply_window_m) == 0);
  assert(pthread_mutex_destroy(&procs_m) == 0);

  wfs_log(WFS_DBG_2, "~rpcs done\n");

}

void
rpcs::setlossy(bool x)
{
  lossy = x;
  if (x)
    chan->setlossy();
  else
    chan->setlossy(0);
}

handler::handler()
{
}

void
rpcs::reg1(unsigned int proc, handler *h)
{
  assert(pthread_mutex_lock(&procs_m) == 0);
  assert(procs.count(proc) == 0);
  procs[proc] = h;
  assert(procs.count(proc) >= 1);
  assert(pthread_mutex_unlock(&procs_m) == 0);
}

void
rpcs::updatestat(unsigned int proc)
{
  counts[proc]++;
  counting--;
  if( counting == 0 ) {
    std::map<int, int>::iterator i;
    printf( "RPC STATS: " );
    for( i = counts.begin(); i != counts.end(); i++ ) {
      printf( "%x:%d ", i->first, i->second );
    }
    printf("\n");
    char *count_env = getenv("RPC_COUNT");
    if(count_env != NULL){
      counting = atoi(count_env);
    }
  }
}

void
rpcs::send(const charbuf cb, int channo, unsigned int xid) {
  // must delete from the in-progress list before sending, in case
  // the channel gets deleted and the list is cleared after the send.
  if( channo ) {
    pthread_mutex_lock(&nt_lock);
    tcp_in_progress[channo].erase(xid);
    pthread_mutex_unlock(&nt_lock);
  }
  chan->send(cb, channo);
}

void
rpcs::dispatch(junk *j)
{
  charbuf xreq = j->s;
  int channo = j->chan;
  delete j;
  unmarshall args(xreq.buf, xreq.sz);

  int clt_nonce, srv_nonce;
  unsigned int proc, xid, rep_xid;
  args.zero_ind();
  args >> clt_nonce;
  args >> srv_nonce;
  args >> proc;
  args >> xid;
  args >> rep_xid;
  assert(args.ind() <= RPC_HEADER_SZ);
  args.reset_ind();

  marshall rep( true );
  charbuf cb1;
  bool old;
  bool cb_present;
  int ret;
  handler *h;

  assert(pthread_mutex_lock(&procs_m) == 0);

  if(!args.ok() || procs.count(proc) < 1){
    assert(pthread_mutex_unlock(&procs_m) == 0);
    wfs_log(WFS_DBG_2, "bad proc %x args ok? %d procs.count %u\n",
		      proc, args.ok()?1:0, (unsigned)procs.count(proc));
    // indicate we won't be replying (to free resources)
    charbuf cb(NULL, 0);
    send(cb, channo, xid);
    goto exit;
  }

  h = procs[proc];

  assert(pthread_mutex_unlock(&procs_m) == 0);

  wfs_log(WFS_DBG_2, "rpc %u (proc %x, last_rep %u) from clt %u for srv instance %u \n",
	    xid, proc, rep_xid, clt_nonce, srv_nonce);

  assert(pthread_mutex_lock(&reply_window_m) == 0);
  // if we don't know about this clt_nonce, create a cleanup object
  if (reply_window.find(clt_nonce) == reply_window.end()) {
    assert (reply_window[clt_nonce].size() == 0); // create
    wfs_log(WFS_DBG_2, "New reply window size is %d\n", reply_window.size() );
    schan::cleanup *obj = new rpcs::cleanup(this, clt_nonce);
    wfs_log(WFS_DBG_2, "Setting cleanup obj for channo %d and clt_nonce %u\n",
		       channo, clt_nonce );
    if (!chan->set_cleanup(channo, obj)) {
      delete obj;
    }
    assert (reply_window.find(clt_nonce) != reply_window.end());
  }
  assert(pthread_mutex_unlock(&reply_window_m) == 0);

  if (nonce != 0 && srv_nonce != 0 && srv_nonce != nonce) {
    wfs_log(WFS_DBG_2, "rpc for an old server instance %u (current %u) proc %x\n",
	     srv_nonce, nonce, proc);
    rep.zero_ind();
    rep << xid;
    rep << rpc_const::atmostonce_failure;
#ifdef VIVALDI
      if (proc & rpc_const::VIVALDI_MASK) {
	assert(_vivaldi);
	rep << _vivaldi->my_coords();
      } else {
	rep << 0 << 0 << 0; // init memory
      }
#endif
    rep.reset_ind();
    int cbsize = rep.size();
    charbuf cb = charbuf(rep.take_buf(), cbsize);
    cb.free = true;
    send(cb, channo, xid);
    goto exit;
  }

  if (checkduplicate_and_update(clt_nonce, xid, channo, rep_xid, old,
				cb_present, cb1)) {
    if (old) { // very old request; don't have response anymore
      wfs_log(WFS_DBG_2, "very old request %u from %u\n", xid, clt_nonce);
      rep.zero_ind();
      rep << xid;
      rep << rpc_const::atmostonce_failure;
#ifdef VIVALDI
      if (proc & rpc_const::VIVALDI_MASK) {
	assert(_vivaldi);
	rep << _vivaldi->my_coords();
      } else {
	rep << 0 << 0 << 0; // init memory
      }
#endif
      rep.reset_ind();
      int cbsize = rep.size();
      charbuf cb = charbuf(rep.take_buf(), cbsize);
      cb.free = true;
      send(cb, channo, xid);
    } else if (cb_present) { // duplicate and we still have the response
      send(cb1, channo, xid);
    } else {
      // server is working on this request
      // indicate we won't be replying (to free resources)
      charbuf cb(NULL, 0);
      send(cb, channo, xid);
    }
    goto exit;
  }

  if (counting) {
    updatestat(proc);
  }

  // This is where we validate the identity, sorta
  bool good;
  if (h->uses_ident()) {
    // We see if the user is authorized to say they are who they
    // say they are are; authorization for the user they say they are
    // to do stuff is in the server code
    std::string ident;
    args >> ident;
    args.reset_ind();
    good = chan->verify_user(ident.c_str(),channo);
  } else if (!h->allow_users()) {
    // Only host can call these methods
    good = chan->verify_user("host",channo);
  } {
    // No need to call verify_user if we're a function that does not use
    // ident and that any user can call; this is authorization, not
    // authentication, and authentication only lets existant users in
    good = true;
  }

  if (good)
    ret = h->fn(args, rep);
  else
    ret = rpc_const::identity_failure;

  rep.zero_ind();
  rep << xid;
  rep << ret;
#ifdef VIVALDI
  if (proc & rpc_const::VIVALDI_MASK) {
    assert(_vivaldi);
    rep << _vivaldi->my_coords();
  } else {
    rep << 0 << 0 << 0; // init memory
  }
#endif

  rep.reset_ind();

  {
    int cbsize = rep.size();
    charbuf cb = charbuf(rep.take_buf(), cbsize);
    cb.refcount = (int *) malloc(sizeof(int));
    *(cb.refcount) = 1;
    wfs_log(WFS_DBG_2, "sending and saving reply of size %d for rpc %u, proc %x, clt %u, checksum %llu\n",
	      cbsize, xid, proc, clt_nonce,
#ifdef WFS_RPC_CHECKSUMMING
	      cb.checksum()
#else
	      (unsigned long long) 0
#endif
	      );

    bool added = add_reply(clt_nonce, xid, cb);
    // TODO: shouldn't bother sending un-added replies in the first place
    if(!added) {
      chan->lock_for_delete(channo);
    }
    send(cb, channo, xid);
    if(!added) {
      cb.decref();
      chan->unlock_for_delete(channo);
    }
  }

 exit:

  pthread_mutex_lock(&nt_lock);
  num_threads[channo]--;
  if( num_threads[channo] == 0 ) {
    num_threads.erase(channo);
  }
  pthread_cond_signal(&nt_cond);
  pthread_mutex_unlock(&nt_lock);
  assert (pthread_cond_signal(&delete_c) == 0);
  pthread_exit(0);

}

void rpcs::do_cleanup(int channo, int clt_nonce) {

  // delete all state associated with this channel
  pthread_mutex_lock(&nt_lock);
  tcp_in_progress.erase(channo);
  pthread_mutex_unlock(&nt_lock);

  wfs_log(WFS_DBG_2,"Doing cleanup for channo %d and clt_nonce %u\n", channo, clt_nonce);

  // reply window stuff
  assert(pthread_mutex_lock(&reply_window_m) == 0);
  std::vector<reply_t *> to_decref;
  std::list<reply_t *>::iterator it = reply_window[clt_nonce].begin();
  while (it != reply_window[clt_nonce].end()) {
    to_decref.push_back(reply_window[clt_nonce].front());
    reply_window[clt_nonce].pop_front();
    it = reply_window[clt_nonce].begin();
  }
  reply_window.erase(clt_nonce);
  assert(pthread_mutex_unlock(&reply_window_m) == 0);

  // do this outside of the reply_window lock, so that it won't block others
  for(unsigned i = 0; i < to_decref.size(); i++ ) {
    reply_t *x = to_decref[i];
    wfs_log(WFS_DBG_2, "Freeing buffer for rpc %d, clnt_nonce %u, channo %d\n",
	      x->xid, clt_nonce, channo );
    chan->lock_for_delete(channo);
    x->cb.decref();
    chan->unlock_for_delete(channo);
    free(x->cb.buf);
    delete x;
  }

  wfs_log(WFS_DBG_2,"Done cleanup for channo %d and clt_nonce %u\n", channo, clt_nonce);

}

bool rpcs::add_reply(unsigned int clt_nonce, unsigned int xid,
		     charbuf &cb)
{
  std::list<reply_t *>::iterator it;
  assert(pthread_mutex_lock(&reply_window_m) == 0);
  bool added = true;

  // remember the reply for this xid.
  for (it = reply_window[clt_nonce].begin();
       it != reply_window[clt_nonce].end(); it++) {
    if ((*it)->xid == xid) {
      (*it)->cb_present = true;
      (*it)->cb = cb;
      goto release;
    }
  }
  //assert(0);  Don't assert this; could be the case that the channel is dead
  // also, the sender could have timed out waiting for the reply and faked
  // an ack for that xid already
  wfs_log(WFS_DBG_2, "Message %u on clt_nonce %d not added to reply queue\n",
	    xid, clt_nonce );
  cb.free = true;
  added = false;

 release:
  assert(pthread_mutex_unlock(&reply_window_m) == 0);
  return added;

}

void
rpcs::free_reply_window(void)
{
  std::map<unsigned int,std::list<reply_t *> >::iterator clt;
  std::list<reply_t *>::iterator it;

  assert(pthread_mutex_lock(&reply_window_m) == 0);
  for (clt = reply_window.begin(); clt != reply_window.end(); clt++) {
    for (it = clt->second.begin(); it != clt->second.end(); it++) {
      // no need to lock, chan has been deleted
      (*it)->cb.decref();
      free((*it)->cb.buf);
      delete *it;
    }
    clt->second.clear();
  }
  reply_window.clear();
  assert(pthread_mutex_unlock(&reply_window_m) == 0);
}

bool rpcs::checkduplicate_and_update(unsigned int clt_nonce, unsigned int xid,
				     int channo, unsigned int xid_rep,
				     bool &old, bool &cb_present, charbuf &cb)
{
  std::list<reply_t *>::iterator it;
  bool r = false;

  assert(pthread_mutex_lock(&reply_window_m) == 0);

  // check if xid is a duplicate, and if not update list of received xid, so
  // that checking and update is atomic.  xid_rep tells srv which replies
  // has received and the server can forget about.
  if (reply_window[clt_nonce].size() > 1 &&
      reply_window[clt_nonce].front()->xid > xid) {
      old = true;
      r = true;
      goto release;
  }

  for (it = reply_window[clt_nonce].begin();
       it != reply_window[clt_nonce].end(); it++) {
    if ((*it)->xid == xid) {
      cb_present = (*it)->cb_present;
      cb = (*it)->cb;
      old = false;
      r = true;
      goto release;
    } else if ((*it)->xid > xid) {
      reply_window[clt_nonce].insert(it, new reply_t(xid));
      goto release;
    }
  }
  reply_window[clt_nonce].push_back(new reply_t(xid));
 release:
  it = reply_window[clt_nonce].begin();
  std::vector<reply_t *> to_decref;
  while (it != reply_window[clt_nonce].end()) {
    if ((*it)->xid > xid_rep) {
      break;
    } else {
      to_decref.push_back(reply_window[clt_nonce].front());
      reply_window[clt_nonce].pop_front();
      it = reply_window[clt_nonce].begin();
    }
  }
  assert(pthread_mutex_unlock(&reply_window_m) == 0);

  // do this outside of the reply_window lock, so that it won't block others
  for(unsigned i = 0; i < to_decref.size(); i++ ) {
    reply_t *x = to_decref[i];
    wfs_log(WFS_DBG_2, "Freeing buffer for rpc %d, clnt_nonce %u, channo %d\n",
	      x->xid, clt_nonce, channo );
    chan->lock_for_delete(channo);
    x->cb.decref();
    chan->unlock_for_delete(channo);
    free(x->cb.buf);
    delete x;
  }

  return r;
}

void
rpcs::loop()
{

  int oldstate;
  assert(pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &oldstate) == 0);

  while(1){
    charbuf req;
    int channo;
    chan->recv(req, channo);

    if(req.sz < 8) {
      // maybe it's a portscan or something on a non-ssh channel
      delete req.buf;
      // indicate we won't be replying (to free resources)
      charbuf cb(NULL, 0);
      chan->send(cb, channo);
      continue;
    }

    // too many threads will make us run out of memory
    // (default linux stack size of a new thread: 10M)
    // TODO: make it a thread pool, so we don't have to
    // keep re-allocating memory?
    pthread_mutex_lock(&nt_lock);

    unmarshall args(req.buf, req.sz);
    unsigned int proc, xid;
    args.zero_ind();
    int clt_nonce, srv_nonce;
    args >> clt_nonce;
    args >> srv_nonce;
    args >> proc;
    args >> xid;
    // take the buffer, because otherwise if this isn't a dup packet
    // the unmarshall will delete the buffer when it goes out of scope
    char *buf = args.take_buf();

    // first make sure this RPC is not already in progress,
    // so we don't waste time and queue space on it
    if( (channo && tcp_in_progress[channo].find(xid) !=
	 tcp_in_progress[channo].end() ) ) {
      wfs_log(WFS_DBG_2, "DROPPING DUP RPC %x %d %d %d\n", proc, xid, channo, clt_nonce);
      free(buf);
      pthread_mutex_unlock(&nt_lock);
      // indicate we won't be replying (to free resources)
      charbuf cb(NULL, 0);
      chan->send(cb, channo);
      continue;
    } else {
      if( channo ) {
	tcp_in_progress[channo][xid] = true;
      }
    }

    if(num_threads[channo] >= PER_CHANNEL_OUTSTANDING){
      wfs_log(WFS_DBG_2, "Too many threads in progress, rejecting packet: %x %d %d %d\n",
          proc, xid, channo, clt_nonce );
      free(buf);
      pthread_mutex_unlock(&nt_lock);
      // indicate we won't be replying (to free resources)
      charbuf cb(NULL, 0);
      send(cb, channo, xid);
      continue;
    }
    num_threads[channo]++;
    pthread_mutex_unlock(&nt_lock);

    junk *j = new junk(req, channo);

    assert (pthread_mutex_lock(&delete_m) == 0);
    if (!deleting) {
      if(method_thread(this, true, &rpcs::dispatch, j) == 0){
	perror("rpcs::loop pthread_create");
	exit(1);
      }
    }
    assert (pthread_mutex_unlock(&delete_m) == 0);
  }
}

int rpcs::bind(int a, int &r)
{
  wfs_log(WFS_DBG_2, "rpcs::bind called nonce %u\n", nonce);
  r = nonce;
  return 0;
}

void
marshall::rawbyte(unsigned char x)
{
  if (_ind >= _capa) {
    _capa *= 2;
    assert (_buf != NULL);
    _buf = (char *)realloc(_buf, _capa);
    assert(_buf);
  }
  _buf[_ind++] = x;
}

void
marshall::rawbytes(const char *p, int n)
{
  if ((_ind+n) > _capa) {
    _capa = _capa > n? 2*_capa:(_capa+n);
    assert (_buf != NULL);
    _buf = (char *)realloc(_buf, _capa);
    assert(_buf);
  }
  memcpy(_buf+_ind, p, n);
  _ind += n;
}

marshall &
operator<<(marshall &m, unsigned char x)
{
  m.rawbyte(x);
  return m;
}

marshall &
operator<<(marshall &m, char x)
{
  m << (unsigned char) x;
  return m;
}


marshall &
operator<<(marshall &m, unsigned short x)
{
  m.rawbyte(x & 0xff);
  m.rawbyte((x >> 8) & 0xff);
  return m;
}

marshall &
operator<<(marshall &m, short x)
{
  m << (unsigned short) x;
  return m;
}

marshall &
operator<<(marshall &m, unsigned int x)
{
  m.rawbyte(x & 0xff);
  m.rawbyte((x >> 8) & 0xff);
  m.rawbyte((x >> 16) & 0xff);
  m.rawbyte((x >> 24) & 0xff);
  return m;
}

marshall &
operator<<(marshall &m, unsigned long x)
{
  m.rawbyte(x & 0xff);
  m.rawbyte((x >> 8) & 0xff);
  m.rawbyte((x >> 16) & 0xff);
  m.rawbyte((x >> 24) & 0xff);
  return m;
}

marshall &
operator<<(marshall &m, int x)
{
  m << (unsigned int) x;
  return m;
}

marshall &
operator<<(marshall &m, const std::string &s)
{
  m << (unsigned int) s.size();
  m.rawbytes(s.data(), s.size());
  return m;
}

marshall &
operator<<(marshall &m, unsigned long long x)
{
  m << (unsigned int) x;
  m << (unsigned int) (x >> 32);
  return m;
}

//take the contents from another unmarshall object
void
unmarshall::take_in(unmarshall &another)
{
  if (_buf)
    free(_buf);
  _sz = another.size();
  _buf = another.take_buf();
  _ind = (_reply?RPC_REPLY_HEADER_SZ:RPC_HEADER_SZ);
  _ok = _sz >= (_reply?RPC_REPLY_HEADER_SZ:RPC_HEADER_SZ)? true:false;
}

bool
unmarshall::okdone()
{
  if(ok() && _ind == _sz) {
    return true;
  } else {
    // an RPC request or reply was too short or too long.
#if 0
    unsigned int *p = (unsigned int *) s.str().data();
    wfs_log(WFS_DBG_OFF, "okdone failed len=%d %x %x %x\n",
            s.str().size(), p[0], p[1], p[2]);
#endif
    return false;
  }
}

unsigned int
unmarshall::rawbyte()
{
  char c = 0;
  if (_ind >= _sz)
    _ok = false;
  else
    c = _buf[_ind++];
  return c;
}

unmarshall &
operator>>(unmarshall &u, unsigned char &x)
{
  x = (unsigned char) u.rawbyte() ;
  return u;
}

unmarshall &
operator>>(unmarshall &u, char &x)
{
  x = (char) u.rawbyte();
  return u;
}


unmarshall &
operator>>(unmarshall &u, unsigned short &x)
{
  x = u.rawbyte() & 0xff;
  x |= (u.rawbyte() & 0xff) << 8;
  return u;
}

unmarshall &
operator>>(unmarshall &u, short &x)
{
  x = u.rawbyte() & 0xff;
  x |= (u.rawbyte() & 0xff) << 8;
  return u;
}

unmarshall &
operator>>(unmarshall &u, unsigned int &x)
{
  x = u.rawbyte() & 0xff;
  x |= (u.rawbyte() & 0xff) << 8;
  x |= (u.rawbyte() & 0xff) << 16;
  x |= (u.rawbyte() & 0xff) << 24;
  return u;
}

unmarshall &
operator>>(unmarshall &u, unsigned long &x)
{
  x = u.rawbyte() & 0xff;
  x |= (u.rawbyte() & 0xff) << 8;
  x |= (u.rawbyte() & 0xff) << 16;
  x |= (u.rawbyte() & 0xff) << 24;
  return u;
}

unmarshall &
operator>>(unmarshall &u, int &x)
{
  x = u.rawbyte() & 0xff;
  x |= (u.rawbyte() & 0xff) << 8;
  x |= (u.rawbyte() & 0xff) << 16;
  x |= (u.rawbyte() & 0xff) << 24;
  return u;
}

unmarshall &
operator>>(unmarshall &u, unsigned long long &x)
{
  unsigned int a, b;
  u >> a;
  u >> b;
  x = a | ((unsigned long long) b << 32);
  return u;
}

unmarshall &
operator>>(unmarshall &u, std::string &s)
{
  unsigned sz;
  u >> sz;
  if(u.ok())
    u.rawbytes(s, sz);
  return u;
}

void
unmarshall::rawbytes(std::string &ss, unsigned int n)
{
  if ((_ind+n) > (unsigned)_sz) {
    _ok = false;
  } else {
    std::string tmps = std::string(_buf+_ind, n);
    swap(ss, tmps);
    assert(ss.size() == n);
    _ind += n;
  }
}

bool operator<(const sockaddr_in &a, const sockaddr_in &b) {
  return ((a.sin_addr.s_addr < b.sin_addr.s_addr) ||
	  ((a.sin_addr.s_addr == b.sin_addr.s_addr) &&
	   ((a.sin_port < b.sin_port))));
}

void
make_sockaddr(const char *hostandport, struct sockaddr_in *dst){

  char host[200];
  const char *localhost = "127.0.0.1";
  const char *port = index(hostandport, ':');
  if(port == NULL){
    memcpy(host, localhost, strlen(localhost)+1);
    port = hostandport;
  }else{
    memcpy(host, hostandport, port-hostandport);
    host[port-hostandport] = '\0';
    port++;
  }

  make_sockaddr(host, port, dst);

}

void
make_sockaddr(const char *host, const char *port, struct sockaddr_in *dst){

  in_addr_t a;

  bzero(dst, sizeof(*dst));
  dst->sin_family = AF_INET;

  a = inet_addr(host);
  if(a != INADDR_NONE){
    dst->sin_addr.s_addr = a;
  } else {
    struct hostent *hp = gethostbyname(host);
    if(hp == 0 || hp->h_length != 4){
      fprintf(stderr, "cannot find host name %s\n", host);
      exit(1);
    }
    dst->sin_addr.s_addr = ((struct in_addr *)(hp->h_addr))->s_addr;
    //fprintf(stderr,"temp sockaddr: doing %s\n",inet_ntoa(dst->sin_addr));
  }
  dst->sin_port = htons(atoi(port));
}

// Set this to a public-key-evaluating function to enable security
static public_key_evaluator pubkeyeval = NULL;
static private_key_source privkeysrc = NULL;
static const char *primuser = NULL;

void rpc_enable_security(public_key_evaluator evaluator,
			 private_key_source source,
			 const char *primary_user) {
  pubkeyeval = evaluator;
  privkeysrc = source;
  primuser = primary_user;
}
bool rpc_is_secure() {
  return pubkeyeval != NULL;
}
bool rpc_public_key_identifies_user(const char *username,
				const char *public_key) {
  return pubkeyeval(username,public_key);
}
const char *rpc_private_key_path(const char *username) {
  return privkeysrc(username);
}
const char *rpc_primary_user() {
  return primuser;
}
