/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#ifndef rpc_h
#define rpc_h

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <list>
#include <map>
#include "marshall.h"
#include "chan.h"

class rpc_const {
  public:
  static const unsigned int bind = 1;   // handler number reserved for bind
  static const int VIVALDI_MASK=0x40000000;
  static const int timeout_failure = -1;
  static const int unmarshal_args_failure = -2;
  static const int unmarshal_reply_failure = -3;
  static const int atmostonce_failure = -4;
  static const int bind_failure = -5;
  static const int cancel_failure = -6;
  static const int identity_failure = -7;
};
// rpc client endpoint.
// manages a socket and an xid space.
// threaded: multiple threads can be sending RPCs,
// and this code gives each reply to the right thread.
class rpcc {
 private:

  const sockaddr_in dst;   // address of server
  unsigned int xid;  // xid for next request/call
  // xids of replies that clnt hasn't acknowledged
  std::list<unsigned int> xid_rep_window;
  int svr_nonce;
  int clt_nonce;
  bool bind_done;
  cchan *chan;
  class vivaldi *_vivaldi;

  // clock loop data for tight timeouts
  pthread_mutex_t _timeout_lock;
  pthread_cond_t _timeout_cond;
  struct timespec _next_timeout;
  static void cleanup_timeout_lock(void *arg) {
    pthread_mutex_t *l = (pthread_mutex_t *) arg;
    pthread_mutex_unlock(l);
  };

  // map xid of awaited reply to waiting thread in call().
  struct caller {
    caller(int xxid, unmarshall *un, uint32_t ip, uint16_t port, double when = 0);
    ~caller();
    int xid;
    unmarshall *un;
    int intret;
    bool done;
    pthread_cond_t c;
    pthread_mutex_t m;
    double senttime;
    uint32_t other_ip;
    uint16_t other_port;
  };
  std::map<int, caller*> calls;
  pthread_mutex_t m; // protect insert/delete to calls[]
  pthread_cond_t destroy_wait_c;
  bool destroy_wait;

  pthread_mutex_t _sender_lock; // make sure sends are sent in the right order

  pthread_t th_chan_loop;
  pthread_t th_clock_loop;
  void chan_loop();
  void clock_loop();
  void got_reply(unmarshall &rep);
  void update_xid_rep(unsigned int xid);

 public:
  rpcc(sockaddr_in _dst);
  ~rpcc();

  bool dead() { return chan->dead(); }

  void setlossy(bool x);
  void set_vivaldi(vivaldi *v) { _vivaldi = v; }

  // hack to allow optional timeout argument to call().
  struct TO {
    int to;
    bool use_tcp;
  };
  static const TO to_inf;
  static TO to(int x, bool use_tcp = false ) { 
    TO t; 
    t.to = x; 
    t.use_tcp = use_tcp;  // expecting a big reply?
    return t;
  }

  int bind(TO to = to_inf);
  int id() { return clt_nonce; }
  void cancel();

  int call1(unsigned int proc, marshall &req, unmarshall &rep, 
	    TO to, const char *required_user);

  template<class R>
    int call_m(unsigned int proc, marshall &req, R & r, TO to, 
	       const char *required_user);

  template<class R>
    int call(unsigned int proc, R & r, TO to = to_inf, 
	     const char *required_user=NULL);
  template<class R, class A1>
    int call(unsigned int proc, const A1 & a1, R & r, TO to = to_inf, 
	     const char *required_user=NULL);
  template<class R, class A1, class A2>
    int call(unsigned int proc, const A1 & a1, const A2 & a2, R & r, 
	     TO to = to_inf, 
	     const char *required_user=NULL);
  template<class R, class A1, class A2, class A3>
    int call(unsigned int proc, const A1 & a1, const A2 & a2, const A3 & a3, 
	     R & r, TO to = to_inf, 
	     const char *required_user=NULL);
  template<class R, class A1, class A2, class A3, class A4>
    int call(unsigned int proc, const A1 & a1, const A2 & a2, const A3 & a3, 
	     const A4 & a4, R & r, TO to = to_inf, 
	     const char *required_user=NULL);
  template<class R, class A1, class A2, class A3, class A4, class A5>
    int call(unsigned int proc, const A1 & a1, const A2 & a2, const A3 & a3, 
	     const A4 & a4, const A5 & a5, R & r, TO to = to_inf, 
	     const char *required_user=NULL);
  template<class R, class A1, class A2, class A3, class A4, class A5,
    class A6>
    int call(unsigned int proc, const A1 & a1, const A2 & a2, const A3 & a3, 
	     const A4 & a4, const A5 & a5, const A6 & a6,
	     R & r, TO to = to_inf, 
	     const char *required_user=NULL);
  template<class R, class A1, class A2, class A3, class A4, class A5, 
    class A6, class A7>
    int call(unsigned int proc, const A1 & a1, const A2 & a2, const A3 & a3, 
	     const A4 & a4, const A5 & a5, const A6 &a6, const A7 &a7,
	     R & r, TO to = to_inf, 
	     const char *required_user=NULL);

};

template<class R> int 
rpcc::call_m(unsigned int proc, marshall &req, R & r, TO to, 
	     const char *required_user)
{
  unmarshall u(true);
  int intret = call1(proc, req, u, to, required_user);
	if (intret < 0) return intret;
  u >> r;
  if(u.okdone() != true)
    return rpc_const::unmarshal_reply_failure;
  return intret;
}

template<class R> int
rpcc::call(unsigned int proc, R & r, TO to, 
	     const char *required_user)
{
  marshall m;
  return call_m(proc, m, r, to, required_user);
}

template<class R, class A1> int
rpcc::call(unsigned int proc, const A1 & a1, R & r, TO to, 
	     const char *required_user)
{
  marshall m;
  m << a1;
  return call_m(proc, m, r, to, required_user);
}

template<class R, class A1, class A2> int
rpcc::call(unsigned int proc, const A1 & a1, const A2 & a2,
           R & r, TO to, 
	   const char *required_user)
{
  marshall m;
  m << a1;
  m << a2;
  return call_m(proc, m, r, to, required_user);
}

template<class R, class A1, class A2, class A3> int
rpcc::call(unsigned int proc, const A1 & a1, const A2 & a2,
           const A3 & a3, R & r, TO to, 
	   const char *required_user)
{
  marshall m;
  m << a1;
  m << a2;
  m << a3;
  return call_m(proc, m, r, to, required_user);
}

template<class R, class A1, class A2, class A3, class A4> int
rpcc::call(unsigned int proc, const A1 & a1, const A2 & a2,
           const A3 & a3, const A4 & a4, R & r, TO to, 
	   const char *required_user)
{
  marshall m;
  m << a1;
  m << a2;
  m << a3;
  m << a4;
  return call_m(proc, m, r, to, required_user);
}

template<class R, class A1, class A2, class A3, class A4, class A5> int
rpcc::call(unsigned int proc, const A1 & a1, const A2 & a2,
           const A3 & a3, const A4 & a4, const A5 & a5, R & r, TO to, 
	   const char *required_user)
{
  marshall m;
  m << a1;
  m << a2;
  m << a3;
  m << a4;
  m << a5;
  return call_m(proc, m, r, to, required_user);
}

template<class R, class A1, class A2, class A3, class A4, class A5,
  class A6> int
rpcc::call(unsigned int proc, const A1 & a1, const A2 & a2,
           const A3 & a3, const A4 & a4, const A5 & a5, 
	   const A6 & a6, R & r, TO to, 
	   const char *required_user)
{
  marshall m;
  m << a1;
  m << a2;
  m << a3;
  m << a4;
  m << a5;
  m << a6;
  return call_m(proc, m, r, to, required_user);
}

template<class R, class A1, class A2, class A3, class A4, class A5,
  class A6, class A7> int
rpcc::call(unsigned int proc, const A1 & a1, const A2 & a2,
           const A3 & a3, const A4 & a4, const A5 & a5, 
	   const A6 & a6, const A7 & a7,
	   R & r, TO to, 
	   const char *required_user)
{
  marshall m;
  m << a1;
  m << a2;
  m << a3;
  m << a4;
  m << a5;
  m << a6;
  m << a7;
  return call_m(proc, m, r, to, required_user);
}

bool operator<(const sockaddr_in &a, const sockaddr_in &b);

class handler {
 public:
  handler();
  virtual ~handler() { }
  virtual int fn(unmarshall &, marshall &) = 0;
  virtual bool uses_ident() {return false;}
  virtual bool allow_users() {return uses_ident();}
};


// rpc server endpoint.
class rpcs {
 private:
  schan *chan;
  bool lossy; // debug: drop some requests and replies
  int lossy_percent; // percentage of packets to drop if lossy is true
  class vivaldi *_vivaldi;
  std::map< int, int > num_threads;
  std::map< int, std::map< int, bool > > tcp_in_progress;
  pthread_mutex_t nt_lock;
  pthread_cond_t nt_cond;

  // deleting rpcs:
  pthread_t th_loop;
  bool deleting;
  pthread_mutex_t delete_m;
  pthread_cond_t delete_c;

  class cleanup : public schan::cleanup {
  private:
    rpcs *_server;
    int _clt_nonce;
  public:
  cleanup(rpcs *server, int clt_nonce) : 
    _server(server), _clt_nonce(clt_nonce) {};
    ~cleanup() {};
    void do_cleanup(int channo) { 
      _server->do_cleanup(channo, _clt_nonce);
      delete this;
    }
  };

  // provide at most once semantics by maintaining a window of replies
  // per client that that client hasn't ackwnowledged receiving yet.
  pthread_mutex_t reply_window_m; // protect reply window et al
  int nonce;
  struct reply_t {
    reply_t (unsigned int _xid) {
      xid = _xid;
      cb_present = false;
    }
    bool cb_present;
    charbuf cb;
    unsigned int xid;
  };
  std::map<unsigned int, std::list<reply_t *> > reply_window;
  void free_reply_window(void);
  bool add_reply(unsigned int clt_nonce, unsigned int xid, charbuf &cb);
  bool checkduplicate_and_update(unsigned int clnt_nonce, 
				 unsigned int xid, int channo, 
				 unsigned int rep_xid,
				 bool &old, bool &cb_present, charbuf &cb);

  // counting
  int counting;
  std::map<int, int> counts;

  // map proc # to function
  std::map<int, handler *> procs;
  pthread_mutex_t procs_m; // protect insert/delete to procs[]
  void updatestat(unsigned int proc);

  void loop();
 protected:
  //void dispatch(std::string, sockaddr_in); // -> compiler error
  struct junk {
    junk(charbuf cb, int c):s(cb),chan(c) {}
    charbuf s;
    int chan;
  };
  void dispatch(junk *);
  void do_cleanup(int channo, int clt_nonce);
  void send(const charbuf cb, int channo, unsigned int xid);

 public:
  rpcs(unsigned int port);
  ~rpcs();

  void set_vivaldi(vivaldi *v) { _vivaldi = v; }

  int bind(int a, int &r);

  unsigned int bound_port() { return chan->bound_port(); }

  // internal handler registration
  void reg1(unsigned int proc, handler *);

  // register a handler
  template<class S, class A1, class R>
    void reg(unsigned int proc, S*, int (S::*meth)(const A1 a1, R & r),
	     bool allow_users=false);
  template<class S, class A1, class A2, class R>
    void reg(unsigned int proc, S*, int (S::*meth)(const A1 a1, const A2, 
						   R & r),
	     bool allow_users=false);
  template<class S, class A1, class A2, class A3, class R>
    void reg(unsigned int proc, S*, int (S::*meth)(const A1, const A2, 
                                          const A3, R & r),
	     bool allow_users=false);
  template<class S, class A1, class A2, class A3, class A4, class R>
    void reg(unsigned int proc, S*, int (S::*meth)(const A1, const A2, 
                                          const A3, const A4, R & r),
	     bool allow_users=false);
  template<class S, class A1, class A2, class A3, class A4, class A5, class R>
    void reg(unsigned int proc, S*, int (S::*meth)(const A1, const A2, 
                                          const A3, const A4, const A5, 
					  R & r),
	     bool allow_users=false);
  template<class S, class A1, class A2, class A3, class A4, class A5, class A6,
    class R>
    void reg(unsigned int proc, S*, int (S::*meth)(const A1, const A2, 
                                          const A3, const A4, const A5, 
					  const A6,
					  R & r),
	     bool allow_users=false);
  template<class S, class A1, class A2, class A3, class A4, class A5, class A6,
    class A7, class R>
    void reg(unsigned int proc, S*, int (S::*meth)(const A1, const A2, 
                                          const A3, const A4, const A5, 
					  const A6, const A7,
					  R & r),
	     bool allow_users=false);

  template<class S, class A1, class R>
    void ident_reg(unsigned int proc, S*, int (S::*meth)(const std::string &i,
							 const A1, R & r));
  template<class S, class A1, class A2, class R>
    void ident_reg(unsigned int proc, S*, int (S::*meth)(const std::string &i, 
							 const A1, const A2, 
						   R & r));
  template<class S, class A1, class A2, class A3, class R>
    void ident_reg(unsigned int proc, S*, int (S::*meth)(const std::string &i, const A1, const A2, 
                                          const A3, R & r));
  template<class S, class A1, class A2, class A3, class A4, class R>
    void ident_reg(unsigned int proc, S*, int (S::*meth)(const std::string &i, const A1, const A2, 
                                          const A3, const A4, R & r));
  template<class S, class A1, class A2, class A3, class A4, class A5, class R>
    void ident_reg(unsigned int proc, S*, int (S::*meth)(const std::string &i, const A1, const A2, 
                                          const A3, const A4, const A5, 
					  R & r));
  template<class S, class A1, class A2, class A3, class A4, class A5, class A6,
    class R>
    void ident_reg(unsigned int proc, S*, int (S::*meth)(const std::string &i, const A1, const A2, 
                                          const A3, const A4, const A5, 
					  const A6,
					  R & r));

  void setlossy(bool);
};

template<class S, class A1, class R> void
rpcs::reg(unsigned int proc, S*sob, int (S::*meth)(const A1 a1, R & r),
	  bool allow_users)
{
  class h1 : public handler {
  private:
    S * sob;
    int (S::*meth)(const A1 a1, R & r);
    bool _users;
  public:
  h1(S *xsob, int (S::*xmeth)(const A1 a1, R & r), bool au)
      : sob(xsob), meth(xmeth), _users(au) { }
    int fn(unmarshall &args, marshall &ret) {
      A1 a1;
      R r;
      args >> a1;
      if(!args.okdone())
        return rpc_const::unmarshal_args_failure;
      int b = (sob->*meth)(a1, r);
      ret << r;
      return b;
    }
    bool allow_users() {return _users;}
  };
  reg1(proc, new h1(sob, meth, allow_users));
}

template<class S, class A1, class A2, class R> void
  rpcs::reg(unsigned int proc, S*sob, int (S::*meth)(const A1 a1, const A2 a2, 
						     R & r),
	  bool allow_users)
{
  class h1 : public handler {
  private:
    S * sob;
    int (S::*meth)(const A1 a1, const A2 a2, R & r);
    bool _users;
  public:
  h1(S *xsob, int (S::*xmeth)(const A1 a1, const A2 a2, R & r), bool au)
    : sob(xsob), meth(xmeth), _users(au)  { }
    int fn(unmarshall &args, marshall &ret) {
      A1 a1;
      A2 a2;
      R r;
      args >> a1;
      args >> a2;
      if(!args.okdone())
        return rpc_const::unmarshal_args_failure;
      int b = (sob->*meth)(a1, a2, r);
      ret << r;
      return b;
    }
    bool allow_users() {return _users;}
  };
  reg1(proc, new h1(sob, meth, allow_users));
}

template<class S, class A1, class A2, class A3, class R> void
  rpcs::reg(unsigned int proc, S*sob, int (S::*meth)(const A1 a1, const A2 a2, 
						     const A3 a3, R & r),
	  bool allow_users)
{
  class h1 : public handler {
  private:
    S * sob;
    int (S::*meth)(const A1 a1, const A2 a2, const A3 a3, R & r);
    bool _users;
  public:
  h1(S *xsob, int (S::*xmeth)(const A1 a1, const A2 a2, const A3 a3, R & r), bool au)
    : sob(xsob), meth(xmeth), _users(au)  { }
    int fn(unmarshall &args, marshall &ret) {
      A1 a1;
      A2 a2;
      A3 a3;
      R r;
      args >> a1;
      args >> a2;
      args >> a3;
      if(!args.okdone())
	return rpc_const::unmarshal_args_failure;
      int b = (sob->*meth)(a1, a2, a3, r);
      ret << r;
      return b;
    }
    bool allow_users() {return _users;}
  };
  reg1(proc, new h1(sob, meth, allow_users));
}

template<class S, class A1, class A2, class A3, class A4, class R> void
  rpcs::reg(unsigned int proc, S*sob, int (S::*meth)(const A1 a1, const A2 a2, 
						     const A3 a3, const A4 a4, 
						     R & r),
	  bool allow_users)
{
  class h1 : public handler {
  private:
    S * sob;
    int (S::*meth)(const A1 a1, const A2 a2, const A3 a3, const A4 a4, R & r);
    bool _users;
  public:
  h1(S *xsob, int (S::*xmeth)(const A1 a1, const A2 a2, const A3 a3, 
			      const A4 a4, R & r), bool au)
    : sob(xsob), meth(xmeth), _users(au)  { }
    int fn(unmarshall &args, marshall &ret) {
      A1 a1;
      A2 a2;
      A3 a3;
      A4 a4;
      R r;
      args >> a1;
      args >> a2;
      args >> a3;
      args >> a4;
      if(!args.okdone())
        return rpc_const::unmarshal_args_failure;
      int b = (sob->*meth)(a1, a2, a3, a4, r);
      ret << r;
      return b;
    }
    bool allow_users() {return _users;}
  };
  reg1(proc, new h1(sob, meth, allow_users));
}

template<class S, class A1, class A2, class A3, class A4, class A5, class R> void
  rpcs::reg(unsigned int proc, S*sob, int (S::*meth)(const A1 a1, const A2 a2, 
						     const A3 a3, const A4 a4, 
						     const A5 a5, R & r),
	  bool allow_users)
{
  class h1 : public handler {
  private:
    S * sob;
    int (S::*meth)(const A1 a1, const A2 a2, const A3 a3, const A4 a4, 
		   const A5 a5, R & r);
    bool _users;
  public:
  h1(S *xsob, int (S::*xmeth)(const A1 a1, const A2 a2, const A3 a3, 
			      const A4 a4, const A5 a5, R & r), bool au)
    : sob(xsob), meth(xmeth), _users(au) { }
    int fn(unmarshall &args, marshall &ret) {
      A1 a1;
      A2 a2;
      A3 a3;
      A4 a4;
      A5 a5;
      R r;
      args >> a1;
      args >> a2;
      args >> a3;
      args >> a4;
      args >> a5;
      if(!args.okdone())
	return rpc_const::unmarshal_args_failure;
      int b = (sob->*meth)(a1, a2, a3, a4, a5, r);
      ret << r;
      return b;
    }
    bool allow_users() {return _users;}
  };
  reg1(proc, new h1(sob, meth, allow_users));
}

template<class S, class A1, class A2, class A3, class A4, class A5, class A6, class R> void
  rpcs::reg(unsigned int proc, S*sob, int (S::*meth)(const A1 a1, const A2 a2, 
						     const A3 a3, const A4 a4, 
						     const A5 a5, const A6 a6, 
						     R & r),
	  bool allow_users)
{
  class h1 : public handler {
  private:
    S * sob;
    int (S::*meth)(const A1 a1, const A2 a2, const A3 a3, const A4 a4, 
		   const A5 a5, const A6 a6, R & r);
    bool _users;
  public:
  h1(S *xsob, int (S::*xmeth)(const A1 a1, const A2 a2, const A3 a3, 
			      const A4 a4, const A5 a5, const A6 a6, R & r), bool au)
    : sob(xsob), meth(xmeth), _users(au) { }
    int fn(unmarshall &args, marshall &ret) {
      A1 a1;
      A2 a2;
      A3 a3;
      A4 a4;
      A5 a5;
      A6 a6;
      R r;
      args >> a1;
      args >> a2;
      args >> a3;
      args >> a4;
      args >> a5;
      args >> a6;
      if(!args.okdone())
	return rpc_const::unmarshal_args_failure;
      int b = (sob->*meth)(a1, a2, a3, a4, a5, a6, r);
      ret << r;
      return b;
    }
    bool allow_users() {return _users;}
  };
  reg1(proc, new h1(sob, meth, allow_users));
}

template<class S, class A1, class A2, class A3, class A4, class A5, 
  class A6, class A7, class R> void
  rpcs::reg(unsigned int proc, S*sob, int (S::*meth)(const A1 a1, const A2 a2, 
						     const A3 a3, const A4 a4, 
						     const A5 a5, const A6 a6,
						     const A7 a7, R & r),
	  bool allow_users)
{
  class h1 : public handler {
  private:
    S * sob;
    int (S::*meth)(const A1 a1, const A2 a2, const A3 a3, const A4 a4, 
		   const A5 a5, const A6 a6, const A7 a7, R & r);
    bool _users;
  public:
  h1(S *xsob, int (S::*xmeth)(const A1 a1, const A2 a2, const A3 a3, 
			      const A4 a4, const A5 a5, const A6 a6,
			      const A7 a7, R & r), bool au)
    : sob(xsob), meth(xmeth), _users(au) { }
    int fn(unmarshall &args, marshall &ret) {
      A1 a1;
      A2 a2;
      A3 a3;
      A4 a4;
      A5 a5;
      A6 a6;
      A7 a7;
      R r;
      args >> a1;
      args >> a2;
      args >> a3;
      args >> a4;
      args >> a5;
      args >> a6;
      args >> a7;
      if(!args.okdone())
	return rpc_const::unmarshal_args_failure;
      int b = (sob->*meth)(a1, a2, a3, a4, a5, a6, a7, r);
      ret << r;
      return b;
    }
    bool allow_users() {return _users;}
  };
  reg1(proc, new h1(sob, meth, allow_users));
}

/* ident_reg */

template<class S, class A1, class R> void
rpcs::ident_reg(unsigned int proc, S*sob, int (S::*meth)(const std::string &i, const A1 a1, R & r))
{
  class h1 : public handler {
  private:
    S * sob;
    int (S::*meth)(const std::string &i, const A1 a1, R & r);
  public:
    h1(S *xsob, int (S::*xmeth)(const std::string &i, const A1 a1, R & r))
      : sob(xsob), meth(xmeth) { }
    int fn(unmarshall &args, marshall &ret) {
      std::string i;
      A1 a1;
      R r;
      args >> i;
      args >> a1;
      if(!args.okdone())
        return rpc_const::unmarshal_args_failure;
      int b = (sob->*meth)(i, a1, r);
      ret << r;
      return b;
    }
    bool uses_ident() {return true;}
  };
  reg1(proc, new h1(sob, meth));
}

template<class S, class A1, class A2, class R> void
  rpcs::ident_reg(unsigned int proc, S*sob, int (S::*meth)(const std::string &i, const A1 a1, const A2 a2, 
						     R & r))
{
  class h1 : public handler {
  private:
    S * sob;
    int (S::*meth)(const std::string &i, const A1 a1, const A2 a2, R & r);
  public:
  h1(S *xsob, int (S::*xmeth)(const std::string &i, const A1 a1, const A2 a2, R & r))
    : sob(xsob), meth(xmeth) { }
    int fn(unmarshall &args, marshall &ret) {
      std::string i;
      A1 a1;
      A2 a2;
      R r;
      args >> i;
      args >> a1;
      args >> a2;
      if(!args.okdone())
        return rpc_const::unmarshal_args_failure;
      int b = (sob->*meth)(i, a1, a2, r);
      ret << r;
      return b;
    }
    bool uses_ident() {return true;}
  };
  reg1(proc, new h1(sob, meth));
}

template<class S, class A1, class A2, class A3, class R> void
  rpcs::ident_reg(unsigned int proc, S*sob, int (S::*meth)(const std::string &i, const A1 a1, const A2 a2, 
						     const A3 a3, R & r))
{
  class h1 : public handler {
  private:
    S * sob;
    int (S::*meth)(const std::string &i, const A1 a1, const A2 a2, const A3 a3, R & r);
  public:
  h1(S *xsob, int (S::*xmeth)(const std::string &i, const A1 a1, const A2 a2, const A3 a3, R & r))
    : sob(xsob), meth(xmeth) { }
    int fn(unmarshall &args, marshall &ret) {
      std::string i;
      A1 a1;
      A2 a2;
      A3 a3;
      R r;
      args >> i;
      args >> a1;
      args >> a2;
      args >> a3;
      if(!args.okdone())
	return rpc_const::unmarshal_args_failure;
      int b = (sob->*meth)(i, a1, a2, a3, r);
      ret << r;
      return b;
    }
    bool uses_ident() {return true;}
  };
  reg1(proc, new h1(sob, meth));
}

template<class S, class A1, class A2, class A3, class A4, class R> void
  rpcs::ident_reg(unsigned int proc, S*sob, int (S::*meth)(const std::string &i, const A1 a1, const A2 a2, 
						     const A3 a3, const A4 a4, 
						     R & r))
{
  class h1 : public handler {
  private:
    S * sob;
    int (S::*meth)(const std::string &i, const A1 a1, const A2 a2, const A3 a3, const A4 a4, R & r);
  public:
  h1(S *xsob, int (S::*xmeth)(const std::string &i, const A1 a1, const A2 a2, const A3 a3, 
			      const A4 a4, R & r))
    : sob(xsob), meth(xmeth) { }
    int fn(unmarshall &args, marshall &ret) {
      std::string i;
      A1 a1;
      A2 a2;
      A3 a3;
      A4 a4;
      R r;
      args >> i;
      args >> a1;
      args >> a2;
      args >> a3;
      args >> a4;
      if(!args.okdone())
        return rpc_const::unmarshal_args_failure;
      int b = (sob->*meth)(i, a1, a2, a3, a4, r);
      ret << r;
      return b;
    }
    bool uses_ident() {return true;}
  };
  reg1(proc, new h1(sob, meth));
}

template<class S, class A1, class A2, class A3, class A4, class A5, class R> void
  rpcs::ident_reg(unsigned int proc, S*sob, int (S::*meth)(const std::string &i, const A1 a1, const A2 a2, 
						     const A3 a3, const A4 a4, 
						     const A5 a5, R & r))
{
  class h1 : public handler {
  private:
    S * sob;
    int (S::*meth)(const std::string &i, const A1 a1, const A2 a2, const A3 a3, const A4 a4, 
		   const A5 a5, R & r);
  public:
  h1(S *xsob, int (S::*xmeth)(const std::string &i, const A1 a1, const A2 a2, const A3 a3, 
			      const A4 a4, const A5 a5, R & r))
    : sob(xsob), meth(xmeth) { }
    int fn(unmarshall &args, marshall &ret) {
      std::string i;
      A1 a1;
      A2 a2;
      A3 a3;
      A4 a4;
      A5 a5;
      R r;
      args >> i;
      args >> a1;
      args >> a2;
      args >> a3;
      args >> a4;
      args >> a5;
      if(!args.okdone())
	return rpc_const::unmarshal_args_failure;
      int b = (sob->*meth)(i, a1, a2, a3, a4, a5, r);
      ret << r;
      return b;
    }
    bool uses_ident() {return true;}
  };
  reg1(proc, new h1(sob, meth));
}

template<class S, class A1, class A2, class A3, class A4, class A5, class A6, class R> void
  rpcs::ident_reg(unsigned int proc, S*sob, int (S::*meth)(const std::string &i, const A1 a1, const A2 a2, 
						     const A3 a3, const A4 a4, 
						     const A5 a5, const A6 a6, 
						     R & r))
{
  class h1 : public handler {
  private:
    S * sob;
    int (S::*meth)(const std::string &i, const A1 a1, const A2 a2, const A3 a3, const A4 a4, 
		   const A5 a5, const A6 a6, R & r);
  public:
  h1(S *xsob, int (S::*xmeth)(const std::string &i, const A1 a1, const A2 a2, const A3 a3, 
			      const A4 a4, const A5 a5, const A6 a6, R & r))
    : sob(xsob), meth(xmeth) { }
    int fn(unmarshall &args, marshall &ret) {
      std::string i;
      A1 a1;
      A2 a2;
      A3 a3;
      A4 a4;
      A5 a5;
      A6 a6;
      R r;
      args >> i;
      args >> a1;
      args >> a2;
      args >> a3;
      args >> a4;
      args >> a5;
      args >> a6;
      if(!args.okdone())
	return rpc_const::unmarshal_args_failure;
      int b = (sob->*meth)(i, a1, a2, a3, a4, a5, a6, r);
      ret << r;
      return b;
    }
    bool uses_ident() {return true;}
  };
  reg1(proc, new h1(sob, meth));
}


void make_sockaddr(const char *hostandport, struct sockaddr_in *dst);
void make_sockaddr(const char *host, const char *port,
		   struct sockaddr_in *dst);


typedef bool (*public_key_evaluator)(const char *username,
				     const char *public_key);
typedef const char *(*private_key_source)(const char *username);
void rpc_enable_security(public_key_evaluator,
			 private_key_source,
			 const char *primary_user);
bool rpc_is_secure();
bool rpc_public_key_identifies_user(const char *username,
				    const char *public_key);
const char *rpc_private_key_path(const char *username);
const char *rpc_primary_user();

#endif
