/*
 * fs-test-5 /mount1/file /mount2/file
 *
 * tests wfs's versioning of the same file, particularly
 * that incomplete new versions are not visible to other nodes.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <dirent.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>

extern int errno;

int
create1(char *f, char fill, int len)
{
  int fd;
  char *p;

  p = (char *) malloc(len);
  assert(p);
  memset(p, fill, len);

  fd = creat(f, 0666);
  if(fd < 0){
    fprintf(stderr, "fs-test-5: create(%s): %s\n",
            f, strerror(errno));
    exit(1);
  }
  if(write(fd, p, len) != len){
    fprintf(stderr, "fs-test-5: write(%s): %s\n",
            f, strerror(errno));
    exit(1);
  }
  free(p);

  return fd;
}

void
check1(char *f, char fill, int len)
{
  int fd, cc, i;
  char *p;
  struct stat st;

  if(stat(f, &st) < 0){
    fprintf(stderr, "fs-test-5: stat(%s): %s\n",
            f, strerror(errno));
    exit(1);
  }
  if(st.st_size != len){
    fprintf(stderr, "fs-test-5: stat(%s) said len %d, wanted %d\n", 
            f, (int)st.st_size, len);
    exit(1);
  }

  fd = open(f, 0);
  if(fd < 0){
    fprintf(stderr, "fs-test-5: open(%s): %s\n",
            f, strerror(errno));
    exit(1);
  }

  p = malloc(len + 1024*32);
  assert(p);
  errno = 0;
  cc = read(fd, p, len + 1);
  if(cc != len){
    fprintf(stderr, "fs-test-5: read(%s) returned %d expected %d%s%s\n",
            f,
            cc,
            len,
            errno ? ": " : "",
            errno ? strerror(errno) : "");
    exit(1);
  }
  close(fd);

  for(i = 0; i < len; i++){
    if(p[i] != fill){
      fprintf(stderr, "fs-test-5: read 0x%02x expected 0x%02x\n",
              p[i] & 0xff, fill & 0xff);
      exit(1);
    }
  }
  free(p);
}

void
unlink1(char *f)
{
  if(unlink(f) != 0){
    fprintf(stderr, "fs-test-5: unlink(%s): %s\n",
            f, strerror(errno));
    exit(1);
  }
}

int
append1(char *f, int fill, int len)
{
  int fd;
  char *p;

  p = malloc(len);
  assert(p);
  memset(p, fill, len);

  fd = open(f, O_WRONLY|O_APPEND);
  if(fd < 0){
    fprintf(stderr, "fs-test-5: append open(%s): %s\n",
            f, strerror(errno));
    exit(1);
  }
  if(write(fd, p, len) != len){
    fprintf(stderr, "fs-test-5: append write(%s): %s\n",
            f, strerror(errno));
    exit(1);
  }
  free(p);

  return fd;
}

int
main(int argc, char *argv[])
{
  int fd;
  char *f1, *f2;

  if(argc != 3){
    fprintf(stderr, "Usage: fs-test-5 file1 file\n");
    exit(1);
  }
  f1 = argv[1];
  f2 = argv[2];

  setbuf(stdout, 0);

  fd = create1(f1, 'a', 23);
  close(fd);
  check1(f1, 'a', 23);
  check1(f2, 'a', 23);

  fd = create1(f1, 'b', 99);
  check1(f2, 'a', 23);
  close(fd);
  check1(f2, 'b', 99);

  fd = append1(f2, 'b', 10);
  check1(f1, 'b', 99);
  close(fd);
  check1(f1, 'b', 109);

  fd = create1(f2, 'c', 15);
  check1(f1, 'b', 109);
  close(fd);
  check1(f1, 'c', 15);

  printf("fs-test-5: passed all tests\n");
  exit(0);
  return 0;
}

