/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <sstream>
#include <arpa/inet.h>
#include "host.h"

host::host(const host &h) {
  addr = h.addr;
  port = h.port;
  site_id = h.site_id;
}

host::host(const char *hname, const char *pname, std::string s )
  : site_id(s)
{
  in_addr_t a;

  a = inet_addr(hname);
  if(a != INADDR_NONE){
    addr = a;
  } else {
    struct hostent *hp = gethostbyname(hname);
    if(hp == 0 || hp->h_length != 4){
      fprintf(stderr, "cannot find host name %s\n", hname);
      exit(1);
    }
    addr = ((struct in_addr *)(hp->h_addr))->s_addr;
  }
  port = htons(atoi(pname));
}

host::operator std::string()
{
  std::ostringstream ost;
  ost << *this;
  return ost.str();
}

const std::string host::str() const
{
  std::ostringstream ost;
  ost << *this;
  return ost.str();
}

std::ostream &
operator<<(std::ostream &os, const host h)
{
  if( h.addr != 0 ) {
    struct in_addr in;
    in.s_addr = h.addr;
    os << inet_ntoa(in) << ":" << ntohs(h.port);
  } else {
    os << "";
  }
  return os;
}

std::ostream &
operator<<(std::ostream &os, const std::vector<host> v)
{
  for(int i = 0; i < (int)v.size(); i++)
    os << v[i] << " ";
  return os;
}

bool
same_hosts(std::vector<host> a, std::vector<host> b)
{
  if(a.size() != b.size())
    return false;
  for(int i = 0; i < (int)a.size(); i++){
    int j;
    for(j = 0; j < (int)b.size(); j++)
      if(a[i] == b[j])
        break;
    if(j >= (int)b.size())
      return false;
  }
  return true;
}

bool
in_hosts(host h, std::vector<host> v)
{
  for(int i = 0; i < (int) v.size(); i++)
    if(v[i] == h)
      return true;
  return false;
}

std::vector<host>
host_union(std::vector<host> a, std::vector<host> b)
{
  std::vector<host> c;
  for(unsigned i = 0; i < a.size(); i++)
    if(!in_hosts(a[i], c))
      c.push_back(a[i]);
  for(unsigned i = 0; i < b.size(); i++)
    if(!in_hosts(b[i], c))
      c.push_back(b[i]);
  return c;
}

