/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include "wfs_file.h"
#include "wfs_misc.h"
#include <arpa/inet.h>
#include <fcntl.h>
#include <math.h>

//#ifdef __linux__
//#include <openssl/md5.h>
//#elif defined(__APPLE__)
//#include <openssl/md5.h>
//#else
//#include <md5.h>  // requires sys/types.h
//#endif /*__linux__*/

// {cue string, cue_indicator, cue_code}
wfs_cue_map_entry wfs_map_entries[] = {
  {"strict", WFS_CUE_CONSISTENCY, WFS_CUE_STRICT},
  {"Strict", WFS_CUE_CONSISTENCY, WFS_CUE_STRICT},
  {"eventualconsistency", WFS_CUE_CONSISTENCY, WFS_CUE_EVENTUAL},
  {"EventualConsistency", WFS_CUE_CONSISTENCY, WFS_CUE_EVENTUAL},
  {"ec", WFS_CUE_CONSISTENCY, WFS_CUE_EVENTUAL},
  {"writelocal", WFS_CUE_DIR_LOC_POLICY, WFS_CUE_WRITE_LOCAL},
  {"WriteLocal", WFS_CUE_DIR_LOC_POLICY, WFS_CUE_WRITE_LOCAL},
  {"wl", WFS_CUE_DIR_LOC_POLICY, WFS_CUE_WRITE_LOCAL},
  {"keeptogether", WFS_CUE_DIR_LOC_POLICY, WFS_CUE_KEEP_TOGETHER},
  {"KeepTogether", WFS_CUE_DIR_LOC_POLICY, WFS_CUE_KEEP_TOGETHER},
  {"kt", WFS_CUE_DIR_LOC_POLICY, WFS_CUE_KEEP_TOGETHER},
  {"wholefile", WFS_CUE_PREFETCH, WFS_CUE_WHOLEFILE},
  {"WholeFile", WFS_CUE_PREFETCH, WFS_CUE_WHOLEFILE},
  {"wf", WFS_CUE_PREFETCH, WFS_CUE_WHOLEFILE},
  {"notwholefile", WFS_CUE_PREFETCH, WFS_CUE_NOTWHOLEFILE},
  {"NotWholeFile", WFS_CUE_PREFETCH, WFS_CUE_NOTWHOLEFILE},
  {"nwf", WFS_CUE_PREFETCH, WFS_CUE_NOTWHOLEFILE},
  {"hotspot", WFS_CUE_POPULARITY, WFS_CUE_HOTSPOT},
  {"HotSpot", WFS_CUE_POPULARITY, WFS_CUE_HOTSPOT},
  {"hs", WFS_CUE_POPULARITY, WFS_CUE_HOTSPOT},
  {"nothotspot", WFS_CUE_POPULARITY, WFS_CUE_NOTHOTSPOT},
  {"NotHotSpot", WFS_CUE_POPULARITY, WFS_CUE_NOTHOTSPOT},
  {"nhs", WFS_CUE_POPULARITY, WFS_CUE_NOTHOTSPOT},
  {"versioned", WFS_CUE_VERSIONING, WFS_CUE_VERSIONED},
  {"notversioned", WFS_CUE_VERSIONING, WFS_CUE_NOTVERSIONED},
  {"nodes", WFS_CUE_SPECIAL, WFS_CUE_NODES},
  {"Nodes", WFS_CUE_SPECIAL, WFS_CUE_NODES},
  {"sites", WFS_CUE_SPECIAL, WFS_CUE_SITES},
  {"Sites", WFS_CUE_SPECIAL, WFS_CUE_SITES},
  {"version", WFS_CUE_SPECIAL, WFS_CUE_VERSION},
  {"", WFS_CUE_SPECIAL, WFS_CUE_NONE}
};

// {cue string, cue_indicator, cue_code}
wfs_cue_map_entry wfs_map_entries_param[] = {
  {"maxtime", WFS_CUE_MAX_TIME, WFS_CUE_NONE},
  {"MaxTime", WFS_CUE_MAX_TIME, WFS_CUE_NONE},
  {"mt", WFS_CUE_MAX_TIME, WFS_CUE_NONE},
  {"replevel", WFS_CUE_REP_LEVEL, WFS_CUE_NONE},
  {"RepLevel", WFS_CUE_REP_LEVEL, WFS_CUE_NONE},
  {"rl", WFS_CUE_REP_LEVEL, WFS_CUE_NONE},
  {"repsites", WFS_CUE_REP_SITES, WFS_CUE_NONE},
  {"RepSites", WFS_CUE_REP_SITES, WFS_CUE_NONE},
  {"rs", WFS_CUE_REP_SITES, WFS_CUE_NONE},
  {"synclevel", WFS_CUE_SYNC_LEVEL, WFS_CUE_NONE},
  {"SyncLevel", WFS_CUE_SYNC_LEVEL, WFS_CUE_NONE},
  {"sl", WFS_CUE_SYNC_LEVEL, WFS_CUE_NONE},
  {"lease", WFS_CUE_LEASE, WFS_CUE_NONE},
  {"staletime", WFS_CUE_STALE_TIME, WFS_CUE_NONE},
  {"StaleTime", WFS_CUE_STALE_TIME, WFS_CUE_NONE},
  {"version", WFS_CUE_VERSION_PARAM, WFS_CUE_NONE},
  {"id", WFS_CUE_ID, WFS_CUE_NONE},
  {"debug", WFS_CUE_SPECIAL, WFS_CUE_DEBUG},
  {"", WFS_CUE_SPECIAL, WFS_CUE_NONE}
};

// {cue string, cue_indicator, cue_code}
wfs_cue_map_entry wfs_map_entries_strparam[] = {
  {"ident", WFS_CUE_IDENT, WFS_CUE_NONE},
  {"Ident", WFS_CUE_IDENT, WFS_CUE_NONE},
  {"", WFS_CUE_SPECIAL, WFS_CUE_NONE}
};

// {cue_indicator, prinf string, permanent cue?}
wfs_ind_property wfs_ind_properties[] = {
  {WFS_CUE_CONSISTENCY, "c", 1},
  {WFS_CUE_MAX_TIME, "mt", 0},
  {WFS_CUE_REP_LEVEL, "rl", 1},
  {WFS_CUE_SYNC_LEVEL, "sl", 0},
  {WFS_CUE_LEASE, "l", 0},
  {WFS_CUE_STALE_TIME, "st", 0},
  {WFS_CUE_DIR_LOC_POLICY, "lp", 1},
  {WFS_CUE_PREFETCH, "pf", 0},
  {WFS_CUE_REP_SITES, "rs", 1},
  {WFS_CUE_POPULARITY, "pop", 0},
  {WFS_CUE_VERSIONING, "ver", 1},
  {WFS_CUE_ID, "id", 0},
  {WFS_CUE_SPECIAL, "", 0}
};

std::map< std::string, int > wfs_map_indices_param;
std::map< std::string, int > wfs_map_indices_strparam;
std::map< std::string, wfs_cached_cue > wfs_cue_cache;

static inline void init_cue_cache() {
  // Initialize the cue cache and if not already done
  // this is a hack and it might get done more than once on a multi-threaded
  // program, but that shouldn't matter too much

  // parameter-less cues can go straight into the cache
  for( int i = 0; wfs_map_entries[i].cue_name != ""; i++ ) {
    wfs_cached_cue cached_cue;
    cached_cue.ind = wfs_map_entries[i].ind;
    cached_cue.val = wfs_map_entries[i].code;
    wfs_cue_cache["." + wfs_map_entries[i].cue_name] = cached_cue;
  }
  // but cues with parameters are just indexed into an array to start...
  for( int i = 0; wfs_map_entries_param[i].cue_name != ""; i++ ) {
    wfs_map_indices_param[wfs_map_entries_param[i].cue_name] = i;
  }
  for( int i = 0; wfs_map_entries_strparam[i].cue_name != ""; i++ ) {
    wfs_map_indices_strparam[wfs_map_entries_strparam[i].cue_name] = i;
  }
}

wfs_file_id node_id_to_file_id( wfs_node_id node_id ) {

  return node_id;

}

wfs_node_id file_id_to_node_id( wfs_file_id file_id ) {
  return file_id;
}

wfs_file_handle file_id_to_file_handle( wfs_file_id file_id ) {

  wfs_file_handle fh;
  fh.fid = file_id;
  return fh;

}

wfs_file_handle file_id_to_file_handle( wfs_file_id file_id,
					wfs_file_handle dir_handle ) {

  wfs_file_handle fh = dir_handle; // transfer the cues
  fh.fid = file_id;
  return fh;

}

pthread_once_t once_control = PTHREAD_ONCE_INIT;
int search_and_set_cues( std::string name, wfs_file_handle &fh ) {

  int cue_component = 0;

  if( name[0] != '.' ) {
    return 0;
  } else {

    if (wfs_cue_cache.size() == 0 ) {
      pthread_once( &once_control, init_cue_cache );
    }

    if( wfs_cue_cache.find(name) != wfs_cue_cache.end() ) {

      wfs_log(WFS_DBG_2, "Cue %s in cache.\n",name.c_str());
      // check the cache -- no lock needed since only inserts are done so far
      wfs_cached_cue cached_cue = wfs_cue_cache[name];

      if( cached_cue.ind == WFS_CUE_SPECIAL ) {
	if( cached_cue.val == WFS_CUE_NODES ) {
	  fh.is_nodes = true;
	} else if( cached_cue.val== WFS_CUE_SITES ) {
	  fh.is_sites = true;
	} else if( cached_cue.val == WFS_CUE_VERSION ) {
	  fh.is_version = true;
	} else {
	  wfs_log(WFS_DBG_OFF, "Bad special cue %s\n", name.c_str() );
	  exit(-1);
	}
      } else {
	if (cached_cue.strval.length() > 0)
	  fh.set_cue( cached_cue.ind, cached_cue.strval);
	else
	  fh.set_cue( cached_cue.ind, cached_cue.val );
      }

      cue_component = 1;

    } else {
      wfs_log(WFS_DBG_2, "Cue %s /not/ in cache, checking for paramness.\n",name.c_str());

      char cue[64];
      int numval;
      char strval[MAX_IDENT_LENGTH+1];
      // ~xavid: use a field width to prevent buffer overflows;
      //         accept syntax with = as well as two dots.
      //         %63[a-zA-Z0-9_] means scan a string of length up to 63 that
      //         contains only letters, digits, and underscores
      if( sscanf( name.c_str(), ".%d.%63s", &numval, cue ) == 2  ||
	  sscanf( name.c_str(),".%63[a-zA-Z0-9_]=%d", cue, &numval) == 2 ) {

	std::string c( cue );
	if( wfs_map_indices_param.find(c) != wfs_map_indices_param.end() ) {
	  wfs_log(WFS_DBG_2, "Cue %s is a parameterized cue with value %d.\n",
		   name.c_str(), numval);

	  int i = wfs_map_indices_param[c];
	  cue_component = 1;

	  if( wfs_map_entries_param[i].ind != WFS_CUE_SPECIAL ) {

	    // add to cache
	    // TODO: currently the cache can grow unbounded, since we expect
	    // a program to only use a small number of parameter cues
	    // however eventually we might want to set a cap (this will
	    // probably need to involve a mutex for consistency)
	    wfs_cached_cue cached_cue;
	    cached_cue.ind = wfs_map_entries_param[i].ind;
	    cached_cue.val = numval;
	    wfs_cue_cache[name] = cached_cue;
	    // activate
	    fh.set_cue( wfs_map_entries_param[i].ind, numval );

	  } else if( wfs_map_entries_param[i].code == WFS_CUE_DEBUG ) {
	    wfs_set_debug( numval );
	    cue_component = WFS_CUE_DEBUG;
	  }

	}
      }
      if (cue_component == 0
	  && (sscanf( name.c_str(), ".%"QUOTE(MAX_IDENT_LENGTH)"[^"INVALID_IDENT_CHARACTERS"].%63s", strval, cue ) == 2  ||
	      sscanf( name.c_str(),".%63[a-zA-Z0-9_]=%"QUOTE(MAX_IDENT_LENGTH)"[^"INVALID_IDENT_CHARACTERS"]", cue, strval) == 2 )
	  && strval[0] != '\0') {

	// Handle cues with string arguments
	std::string c( cue ),v(strval);
	if( wfs_map_indices_strparam.find(c) != wfs_map_indices_strparam.end() ) {
	  wfs_log(WFS_DBG_2, "Cue %s is a string parameterized cue with value %s.\n",
		   name.c_str(), strval);

	  int i = wfs_map_indices_strparam[c];
	  cue_component = 1;

	  if( wfs_map_entries_strparam[i].ind != WFS_CUE_SPECIAL ) {

	    // add to cache
	    // TODO: currently the cache can grow unbounded, since we expect
	    // a program to only use a small number of parameter cues
	    // however eventually we might want to set a cap (this will
	    // probably need to involve a mutex for consistency)
	    wfs_cached_cue cached_cue;
	    cached_cue.ind = wfs_map_entries_strparam[i].ind;
	    cached_cue.strval = v;
	    wfs_cue_cache[name] = cached_cue;
	    // activate
	    fh.set_cue( wfs_map_entries_strparam[i].ind, strval );

	  }
	}
      }
    }
  }

  wfs_log(WFS_DBG_2, "name: %s, cue_component: %d\n", name.c_str(), cue_component);
  return cue_component;

}

unsigned int calc_block_no( unsigned long long off, unsigned int n,
			    unsigned int &num_blocks ) {

  unsigned int b = (unsigned int) floor(off/WFS_BLOCK_SIZE);
  // the block where the last block of the write will exist . . .
  unsigned int b2 = (unsigned int) floor((off+n-1)/WFS_BLOCK_SIZE);
  if( n == 0 ) {
    b2 = b; // if the offset is exactly on the border . . .
  }
  num_blocks = b2-b+1;
  return b;

}

int offset_within_block( unsigned int block_no,
			 unsigned long long full_offset ) {

  // TODO: if the full_offset is huge, it could be too far outside the
  // block to fit in an int
  return (int) (full_offset - (((unsigned long long) block_no)*
			       ((unsigned long long) WFS_BLOCK_SIZE)));

}

wfs_file_handle::wfs_file_handle()
  : fid(0), which_cues(0), is_nodes(false),
                                     is_sites(false),
				     is_version(false),
				     version(-1, (wfs_node_id)-1) {

  for( int i = 0; i < CUE_NUM_CUES; i++ ) {
    cues[i].val = 0;
    cues[i].depth = 0;
  }
  node.id = 0;
  site = "";
}

wfs_file_handle::cue_slot
wfs_file_handle::slot_from_ind( wfs_cue_indicator ind ) {
  int ret, iind = ind;

  // xavid: Using doubles leads to possible rounding problems
  //static const double log2_div = 1/log10(2.0);
  //return (wfs_file_handle::cue_slot) (log10(ind)*log2_div);
  for (ret = -1; iind != 0; ret++)
    iind >>= 1;
  return (wfs_file_handle::cue_slot)ret;
}

wfs_cue_indicator
wfs_file_handle::ind_from_slot( wfs_file_handle::cue_slot slot ) {
  return (wfs_cue_indicator) (1 << ((int) slot));
}

wfs_cue empty;
wfs_cue wfs_file_handle::get_cue( wfs_cue_indicator ind ) const {

  if( which_cues & ind ) {
    return cues[slot_from_ind(ind)];
  } else {
    return empty;
  }

}

void wfs_file_handle::set_cue( wfs_cue_indicator ind, uint32_t val,
			       int depth ) {

  if( ind == WFS_CUE_REP_LEVEL && val < WFS_REP_LEVEL_MIN ) {
    val = WFS_REP_LEVEL_MIN;
  } else if( ind == WFS_CUE_REP_LEVEL && val > WFS_REP_LEVEL_MAX ) {
    val = WFS_REP_LEVEL_MAX;
  }

  which_cues |= ind;
  cues[slot_from_ind(ind)].val = val;
  cues[slot_from_ind(ind)].depth = depth;

}

// For string cues
void wfs_file_handle::set_cue( wfs_cue_indicator ind,
			       const std::string &strval,
			       int depth ) {

  which_cues |= ind;
  cues[slot_from_ind(ind)].strval = strval;
  cues[slot_from_ind(ind)].depth = depth;

}

void wfs_file_handle::add_depth( int change ) {

  // for each cue slot, check which_cues and do the right thing
  // accordingly
  for( int slot = 0; slot < CUE_NUM_CUES; slot++ ) {
    if( which_cues & ind_from_slot( (cue_slot) slot ) ) {
      uint16_t depth = cues[slot].depth;
      if( depth <= 1 && change < 0 ) {
	which_cues &= ~(ind_from_slot( (cue_slot) slot ));
	cues[slot].depth = 0;
      } else {
	cues[slot].depth += change;
      }
    }
  }

}

void wfs_file_handle::inc_depth() {
  add_depth(1);
}

void wfs_file_handle::dec_depth() {
  add_depth(-1);
}

std::string wfs_file_handle::print() {

  int n = 8192;
  int i = 0;
  char *s = new char[n + 1];

  i += snprintf( s+i, n-i, "fid=%016"PRIx64", cues=%d", fid, (int)which_cues );
  for( int j = 0; wfs_ind_properties[j].ind != WFS_CUE_SPECIAL; j++ ) {
    if( which_cues & wfs_ind_properties[j].ind ) {
      wfs_cue cue = get_cue(wfs_ind_properties[j].ind);
      i += snprintf( s+i, n-i, ", %s=%d/%d", wfs_ind_properties[j].str.c_str(),
		     cue.val, cue.depth );
    }
  }

  //i += snprintf( s+i, n-i, " vvv=%d", (int)get_cue(WFS_CUE_VERSION_PARAM).val );

  if( is_nodes ) {
    i += snprintf( s+i, n-i, ", is_nodes=true" );
  }
  if( is_sites ) {
    i += snprintf( s+i, n-i, ", is_sites=true" );
  }
  if( node.id != 0 ) {
    i += snprintf( s+i, n-i, ", node=%s:%d", inet_ntoa( node.addr.sin_addr ),
		   htons(node.addr.sin_port) );
  }

  std::string ss(s);
  delete [] s;
  return ss;

}

// 8 digit cue indicator + '=' + 8 digit hex value + '/' + 4 digit hex depth
static const char *PERM_CUE_STR = "%08x=%08x/%04x";
static const int PERM_CUE_SIZE = 22;

std::string wfs_file_handle::perm_cues_string() {

  // TODO: this buf is too big, but I don't feel like counting the
  // max number of permanent cues . . .
  char buf[CUE_NUM_CUES*PERM_CUE_SIZE+1];
  buf[0] = '\0';
  char *curr_buf = buf;

  for( int i = 0; wfs_ind_properties[i].ind != WFS_CUE_SPECIAL; i++ ) {
    if( wfs_ind_properties[i].perm &&
	(which_cues & wfs_ind_properties[i].ind) ) {
      wfs_cue cue = get_cue(wfs_ind_properties[i].ind);
      sprintf( curr_buf, PERM_CUE_STR, wfs_ind_properties[i].ind, cue.val,
	       cue.depth );
      curr_buf += PERM_CUE_SIZE;
    }
  }

  std::string s( buf );
  return s;

}

void wfs_file_handle::set_perm_cues( std::string perm_cues ) {

  const char *buf = perm_cues.c_str();
  char *curr_buf = (char *) buf;
  while( curr_buf + PERM_CUE_SIZE <= buf + perm_cues.length() ) {
    wfs_cue_indicator ind;
    wfs_cue_code val;
    int depth;
    if( sscanf( curr_buf, PERM_CUE_STR, &ind, &val, &depth ) == 3 ) {
      // TODO: confirm these indicators are perm.
      set_cue( ind, val, (uint16_t) depth );
      curr_buf += PERM_CUE_SIZE;
    }
  }

}

bool operator==(const wfs_version &a, const wfs_version &b) {
    return (a.version_num == b.version_num) && (a.node_id == b.node_id);
}

bool operator!=(const wfs_version &a, const wfs_version &b) {
    return (a.version_num != b.version_num) || (a.node_id != b.node_id);
}

bool operator<(const wfs_version &a, const wfs_version &b) {
    return (a.version_num < b.version_num) ||
        ((a.version_num == b.version_num) && (a.node_id < b.node_id));
}

bool operator<=(const wfs_version &a, const wfs_version &b) {
    return (a.version_num < b.version_num) ||
        ((a.version_num == b.version_num) && (a.node_id <= b.node_id));
}

bool operator>=(const wfs_version &a, const wfs_version &b) {
    return (a.version_num > b.version_num) || ((a.version_num == b.version_num) && (a.node_id >= b.node_id));

}

bool operator>(const wfs_version &a, const wfs_version &b) {
    return (a.version_num > b.version_num) || ((a.version_num == b.version_num) && (a.node_id > b.node_id));

}


bool operator==( const wfs_cue &a, const wfs_cue &b ) {
  return (a.val == b.val) && (a.depth && b.depth) && (a.strval == b.strval);
}

bool operator!=( const wfs_cue &a, const wfs_cue &b ) {
  return !(a == b);
}

bool operator==( const wfs_file_handle &a, const wfs_file_handle &b ) {

  if( a.fid != b.fid || a.which_cues != b.which_cues ||
      a.is_nodes != b.is_nodes || a.node.id != b.node.id ||
      a.is_sites != b.is_sites || a.site != b.site ||
      a.is_version != b.is_version || a.version != b.version ) {
    return false;
  }

  for( int i = 0; i < wfs_file_handle::CUE_NUM_CUES; i++ ) {
    if( i != wfs_file_handle::CUE_SLOT_ID &&
	((a.cues[i].val != b.cues[i].val)
	 || (a.cues[i].strval != b.cues[i].strval))) {
      return false;
    }
  }

  return true;

}

std::string wfs_file_filename( std::string dir, wfs_file_id fid,
			       int32_t vernum) {
  if (vernum <= 0) {
    char buf[WFS_FID_NUM_HEX_BYTES+3];
    sprintf( buf, "/%016"PRIx64"", fid );
    return dir + std::string( buf, strlen(buf) );
  } else {
    char buf[WFS_FID_NUM_HEX_BYTES+3+1+WFS_INT_NUM_DEC_BYTES];
    snprintf(buf,72,
	     "/%016"PRIx64"@%d", fid, vernum);
    return dir + std::string( buf, strlen(buf) );
  }
}

// returns in milliseconds
rpcc::TO fhto( wfs_file_handle fh ) {
  int32_t to_val = fh.get_cue( WFS_CUE_MAX_TIME ).val;
  if( to_val != WFS_CUE_NONE ) {
    return rpcc::to(to_val);
  } else {
    return WFS_DEFAULT_TO;
  }
}

// returns in seconds
int fhlease( wfs_file_handle fh ) {
  int32_t to_val = fh.get_cue( WFS_CUE_LEASE ).val;
  //  if( to_val != WFS_CUE_NONE ) {
  if( fh.which_cues & WFS_CUE_LEASE ) {
    return to_val/1000;
  } else {
    return WFS_LEASE_SEC_DEFAULT;
  }
}

// returns in milliseconds
int fhstaletime( wfs_file_handle fh ) {
  return fh.get_cue( WFS_CUE_STALE_TIME ).val;
}

int fhreplevel( wfs_file_handle fh ) {
  int32_t rl_val = fh.get_cue( WFS_CUE_REP_LEVEL ).val;
  if( rl_val != WFS_CUE_NONE ) {
    return rl_val;
  } else {
    return WFS_REP_LEVEL_DEFAULT;
  }
}

int fhrepsites( wfs_file_handle fh ) {
  int32_t rl_val = fhreplevel(fh);
  int32_t rs_val = fh.get_cue( WFS_CUE_REP_SITES ).val;
  if( rs_val != WFS_CUE_NONE && rs_val <= rl_val ) {
    return rs_val;
  } else {
    return rl_val;
  }
}

int fhsynclevel( wfs_file_handle fh ) {
  int32_t rl_val = fhreplevel(fh);
  int32_t sl_val = fh.get_cue( WFS_CUE_SYNC_LEVEL ).val;
  if( sl_val != WFS_CUE_NONE && sl_val <= rl_val ) {
    return sl_val;
  } else {
    return rl_val;
  }
}

bool wfs_isfile( wfs_file_id fid ) {
  return (fid & WFS_FILE_MASK);
}

bool wfs_islink( wfs_file_id fid ) {
  return (fid & WFS_LINK_MASK);
}

bool wfs_isdir( wfs_file_id fid ) {
  return !wfs_isfile(fid);
}

bool wfs_isfile( wfs_file_handle fh ) {
  return wfs_isfile(file_handle_to_file_id(fh));
}

bool wfs_isdir( wfs_file_handle fh ) {
  return wfs_isdir(file_handle_to_file_id(fh));
}

bool wfs_islink( wfs_file_handle fh ) {
  return wfs_islink(file_handle_to_file_id(fh));
}

std::string format_permissions(const std::map<std::string,
				      wfs_permission> &permissions){
  std::string ret;

  std::map<std::string, wfs_permission>::const_iterator i;
  for (i = permissions.begin(); i != permissions.end(); i++) {
    if (ret.length() > 0)
      ret += ',';
    ret += i->first;
    ret += '=';
    if (i->second & WFS_PERMISSION_ADMIN)
      ret += 'a';
    else
      ret += '-';
    if (i->second & WFS_PERMISSION_READ)
      ret += 'r';
    else
      ret += '-';
    if (i->second & WFS_PERMISSION_WRITE)
      ret += 'w';
    else
      ret += '-';
    if (i->second & WFS_PERMISSION_EXECUTE)
      ret += 'x';
    else
      ret += '-';
  }
  if (ret.length() <= 0)
    return "(none)";
  else
    return ret;
}
