/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#ifndef __WFS_FILE_H__
#define __WFS_FILE_H__ 1

// Get macros for printf/scanfing fixed-size integral types
#ifndef __STDC_FORMAT_MACROS
#define __STDC_FORMAT_MACROS
#endif
#include <ext/hash_map>
#include <inttypes.h>
#include <map>
#include <rpc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include "wfs_node.h"
#include "util/ident_util.h"

using __gnu_cxx::hash_map;

//#define MAIL_HACK 1

typedef uint64_t wfs_file_id;

// cues

enum wfs_cue_code {
  WFS_CUE_NONE = -1,
  WFS_CUE_STRICT = 1,
  WFS_CUE_EVENTUAL,
  WFS_CUE_KEEP_TOGETHER,
  WFS_CUE_WRITE_LOCAL,
  WFS_CUE_WHOLEFILE,
  WFS_CUE_NOTWHOLEFILE,
  WFS_CUE_NOTHOTSPOT,
  WFS_CUE_HOTSPOT,
  WFS_CUE_NOTVERSIONED,
  WFS_CUE_VERSIONED,
  WFS_CUE_NODES,
  WFS_CUE_SITES,
  WFS_CUE_VERSION,
  WFS_CUE_DEBUG,
  WFS_NUM_CUE_CODES
};

typedef struct wfs_cue {

  int32_t val;
  std::string strval;
  uint16_t depth;
  wfs_cue() : val(WFS_CUE_NONE), depth(0) {}

} wfs_cue;

enum wfs_cue_indicator {
  WFS_CUE_SPECIAL = 0,
  WFS_CUE_CONSISTENCY = 0x0001,
  WFS_CUE_MAX_TIME = 0x0002,
  WFS_CUE_REP_LEVEL = 0x0004,
  WFS_CUE_SYNC_LEVEL = 0x0008,
  WFS_CUE_LEASE = 0x0010,
  WFS_CUE_STALE_TIME = 0x0020,
  WFS_CUE_DIR_LOC_POLICY = 0x0040,
  WFS_CUE_PREFETCH = 0x0080,
  WFS_CUE_REP_SITES = 0x0100,
  WFS_CUE_POPULARITY = 0x0200,
  WFS_CUE_VERSIONING = 0x0400,
  WFS_CUE_VERSION_PARAM = 0x0800,
  WFS_CUE_ID = 0x1000,
  WFS_CUE_IDENT = 0x2000
};
// If you add something new here, be sure to add a slot for it as well

typedef struct wfs_cue_map_entry {
  std::string cue_name;
  wfs_cue_indicator ind;
  wfs_cue_code code;
} wfs_cue_map_entry;

typedef struct wfs_cached_cue {
  wfs_cue_indicator ind;
  int32_t val;
  std::string strval; // For cues with string values
} wfs_cached_cue;

typedef struct wfs_ind_property {
  wfs_cue_indicator ind;
  std::string str;
  bool perm;
} wfs_ind_property;

//file version numbers
typedef struct wfs_version {
    int32_t version_num;
    wfs_node_id node_id;

    wfs_version() {version_num = 0; node_id = 0;};
    wfs_version(int32_t _ver, wfs_node_id _id) {
      version_num = _ver; node_id = _id;
    };
} wfs_version;

namespace __gnu_cxx {
  template<>
    class hash<wfs_file_id> {
  public:
    size_t operator()(const wfs_file_id &r) const {
      return (unsigned int32_t) r;
    };
  };
};

// TODO: Problems with file handles:
//  * not clear how to set a cue once it decrements to 0.  Would require
//    keeping a list somewhere of past cues

typedef struct wfs_file_handle {

  enum cue_slot {
    CUE_SLOT_CONSISTENCY = 0,
    CUE_SLOT_MAX_TIME,
    CUE_SLOT_REP_LEVEL,
    CUE_SLOT_SYNC_LEVEL,
    CUE_SLOT_LEASE,
    CUE_SLOT_STALE_TIME,
    CUE_SLOT_DIR_LOC_POLICY,
    CUE_SLOT_PREFETCH,
    CUE_SLOT_REP_SITES,
    CUE_SLOT_POPULARITY,
    CUE_SLOT_VERSIONING,
    CUE_SLOT_VERSION_PARAM,
    CUE_SLOT_ID,
    CUE_SLOT_IDENT,
    CUE_NUM_CUES
  };

  wfs_file_id fid;
  uint16_t which_cues;
  wfs_cue cues[CUE_NUM_CUES];

  bool is_nodes;
  bool is_sites;
  bool is_version;
  wfs_node node;
  std::string site;
  wfs_version version;

  wfs_file_handle();

  static cue_slot slot_from_ind( wfs_cue_indicator ind );
  static wfs_cue_indicator ind_from_slot( wfs_file_handle::cue_slot slot );
  wfs_cue get_cue( wfs_cue_indicator ind ) const;
  void set_cue( wfs_cue_indicator ind, uint32_t val, int depth = 1 );
  void set_cue( wfs_cue_indicator ind, const std::string &strval, 
		int depth = 1 );
  void add_depth( int change );
  void inc_depth();
  void dec_depth();
  std::string print();
  std::string perm_cues_string();
  void set_perm_cues( std::string perm_cues );

} wfs_file_handle;

//comparison operators for wfs version numbers
bool operator==(const wfs_version &a, const wfs_version &b);
bool operator!=(const wfs_version &a, const wfs_version &b);
bool operator<(const wfs_version &a, const wfs_version &b);
bool operator<=(const wfs_version &a, const wfs_version &b);
bool operator>=(const wfs_version &a, const wfs_version &b);
bool operator>(const wfs_version &a, const wfs_version &b);
bool operator==( const wfs_cue &a, const wfs_cue &b );
bool operator==( const wfs_file_handle &a, const wfs_file_handle &b );

uint64_t const WFS_FILE_MASK = 0x0000000000000002;
uint64_t const WFS_LINK_MASK = 0x0000000000000004;
// +xavid
enum wfs_file_type {
  WFS_TYPE_DIRECTORY,
  WFS_TYPE_REGULAR,
  WFS_TYPE_LINK
};

unsigned int const WFS_BLOCK_SIZE = 64*1024;
//unsigned int const WFS_BLOCK_SIZE = 256*1024;
unsigned int const WFS_REP_LEVEL_DEFAULT = 3;
unsigned int const WFS_REP_LEVEL_MIN = 1;
unsigned int const WFS_REP_LEVEL_MAX = 10;
unsigned int const WFS_LEASE_SEC_DEFAULT = 60;
unsigned int const WFS_VERSION_FILE_SIZE = 10;
#define WFS_DIRENTRY_SEP '\t'
#define WFS_DIRENTRY_SEP_Q "\t"
unsigned int const WFS_FID_NUM_HEX_BYTES = sizeof(wfs_file_id)*2;
unsigned int const WFS_DIRENTRY_MAX_NAMELEN = 1024;
unsigned int const WFS_CHECKSUM_LEN = 32;
unsigned int const WFS_INT_NUM_DEC_BYTES = 10;
unsigned int const WFS_INT_NUM_HEX_BYTES = sizeof(uint32_t)*2;
unsigned int const WFS_DIRENTRY_MAX_LEN = 
  WFS_DIRENTRY_MAX_NAMELEN + 1 + WFS_FID_NUM_HEX_BYTES + 2;
const std::string WFS_RANDOM_SITE("random");

const rpcc::TO WFS_DEFAULT_TO = { 10000, false };

wfs_file_id node_id_to_file_id( wfs_node_id node_id );
wfs_node_id file_id_to_node_id( wfs_file_id file_id );

inline
wfs_file_id file_handle_to_file_id( wfs_file_handle file_handle ) {

  return file_handle.fid;
  //return (wfs_file_id) (file_handle >> 32);
  //return file_handle;
}

inline
wfs_file_id two_ints_to_file_id( uint32_t fid_1, uint32_t fid_2 ) {
  wfs_file_id fid = fid_1;
  fid <<= sizeof(uint32_t)*8;
  fid += fid_2;
  return fid;
}

rpcc::TO fhto( wfs_file_handle fh );
int fhlease( wfs_file_handle fh );
int fhstaletime( wfs_file_handle fh );
int fhreplevel( wfs_file_handle fh );
int fhrepsites( wfs_file_handle fh );
int fhsynclevel( wfs_file_handle fh );
wfs_file_handle file_id_to_file_handle( wfs_file_id file_id );
wfs_file_handle file_id_to_file_handle( wfs_file_id file_id, 
					wfs_file_handle dir_handle );
int search_and_set_cues( std::string name, wfs_file_handle &fh );
unsigned int calc_block_no( unsigned long long off, unsigned int n, 
			    unsigned int &num_blocks );
int offset_within_block( unsigned int block_no, 
			 unsigned long long full_offset );

std::string wfs_file_filename( std::string dir, wfs_file_id fid,
			       int32_t vernum = 0);
bool wfs_isfile( wfs_file_handle fh );
bool wfs_isdir( wfs_file_handle fh );
bool wfs_islink( wfs_file_handle fh );
bool wfs_isfile( wfs_file_id fid );
bool wfs_isdir( wfs_file_id fid );
bool wfs_islink( wfs_file_id fid );


// marshalling

inline unmarshall &
operator>>(unmarshall &u, wfs_version &ver)
{
  u >> ver.version_num;
  u >> ver.node_id;
  return u;
}

inline marshall &
operator<<(marshall &m, wfs_version ver) 
{
  m << ver.version_num;
  m << ver.node_id;
  return m;
}

inline unmarshall &
operator>>(unmarshall &u, wfs_cue &cue)
{
  u >> cue.val;
  u >> cue.strval;
  u >> cue.depth;
  return u;
}

inline marshall &
operator<<(marshall &m, wfs_cue cue)
{
  m << cue.val;
  m << cue.strval;
  m << cue.depth;
  return m;
}

inline unmarshall &
operator>>(unmarshall &u, wfs_file_handle &fh)
{
  u >> fh.fid;
  u >> fh.which_cues;
  wfs_cue tmp_cue;
  for( int slot = 0; slot < wfs_file_handle::CUE_NUM_CUES; slot++ ) {
    wfs_cue_indicator ind = 
      wfs_file_handle::ind_from_slot( (wfs_file_handle::cue_slot) slot );
    if( fh.which_cues & ind ) {
      u >> tmp_cue;
      if (tmp_cue.strval.length() > 0)
	fh.set_cue( ind, tmp_cue.strval, tmp_cue.depth);
      else
	fh.set_cue( ind, tmp_cue.val, tmp_cue.depth );
    }
  }
  return u;
}

inline marshall &
operator<<(marshall &m, wfs_file_handle fh)
{
  m << fh.fid;
  m << fh.which_cues;
  for( int slot = 0; slot < wfs_file_handle::CUE_NUM_CUES; slot++ ) {
    wfs_cue_indicator ind = 
      wfs_file_handle::ind_from_slot( (wfs_file_handle::cue_slot) slot );
    if( fh.which_cues & ind ) {
      m << fh.get_cue(ind);
    }
  }
  return m;
}

// xavid: permissions stuff
typedef uint32_t wfs_permission;
const wfs_permission WFS_PERMISSION_ADMIN = 0x0008; // i.e., file owner
const wfs_permission WFS_PERMISSION_READ = 0x0004;
const wfs_permission WFS_PERMISSION_WRITE = 0x0002;
const wfs_permission WFS_PERMISSION_EXECUTE = 0x0001;

const wfs_permission WFS_PERMISSION_ALL = (WFS_PERMISSION_READ
					   | WFS_PERMISSION_WRITE
					   | WFS_PERMISSION_EXECUTE);

const std::string ANYONE_IDENT("anyuser");
const std::string ROOT_IDENT("host");

std::string format_permissions(const std::map<std::string, 
				      wfs_permission> &permissions);

inline bool has_permission(const std::map<std::string, wfs_permission> &permmap,
			   const std::string &user,
			   wfs_permission bits) {
  return user == ROOT_IDENT 
    || (permmap.find(ANYONE_IDENT) != permmap.end() 
	&& (permmap.find(ANYONE_IDENT)->second & bits))
    || (permmap.find(user) != permmap.end() 
	&& (permmap.find(user)->second & bits));
}

// Eventually, this will be a structure used to prove that a user authorized
// an operation.
//typedef int wfs_authorization;

#endif // __WFS_FILE_H__


