/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <assert.h>
#include <utime.h>
#ifdef __linux__
#include <openssl/md5.h>
#elif defined(__APPLE__)
#include <openssl/md5.h>
#else
#include <md5.h>  // requires sys/types.h
#endif /*__linux__*/

#include "wfs_metadata_xattr.h"
#include "wfs_misc.h"
#include "wfs_storage.h"

#define MEMORY_ONLY false
#define MD_XATTR false

wfs_storage::wfs_storage( std::string dir, bool set ) :
  _md_memory_only(MEMORY_ONLY),
  _md_xattr(MD_XATTR) {
  assert( pthread_mutex_init( &_md_lock, NULL ) == 0 );
  set_dir( dir, set );
}

wfs_storage::~wfs_storage() {
  pthread_mutex_destroy( &_md_lock );
}

std::string wfs_storage::filename( wfs_file_handle fh, int32_t vernum ) {
  return wfs_file_filename( _storage_dir, file_handle_to_file_id(fh), vernum );
}

void wfs_storage::set_dir( std::string dir, bool set ) {
  _storage_dir = dir;
  if( set ) {
    wfs_metadata::set_dir(dir);
  }
}

wfs_metadata *wfs_storage::get_md( wfs_file_id fid, int32_t vernum ) {

  if( _md_memory_only ) {
    pthread_mutex_lock( &_md_lock );
    if( _metadata.find(fid) == _metadata.end() ) {
      if( _md_xattr ) {
	_metadata[fid] = new wfs_metadata_xattr( fid, vernum );
      } else {
	_metadata[fid] = new wfs_metadata( fid, vernum );
      }
    }
    wfs_metadata *md = _metadata[fid];
    pthread_mutex_unlock( &_md_lock );
    return md;
  } else {
    if( _md_xattr ) {
      return new wfs_metadata_xattr( fid, vernum );
    } else {
      return new wfs_metadata( fid, vernum );
    }
  }

}

void wfs_storage::release_md( wfs_metadata *md ) {
  if( !_md_memory_only ) {
    delete md;
  }
}

void wfs_storage::set_and_release_md( wfs_file_id fid, wfs_metadata *md ) {

  if( _md_memory_only ) {
    pthread_mutex_lock( &_md_lock );
    _metadata[fid] = md;
    pthread_mutex_unlock( &_md_lock );
  } else {
    bool res = md->write_to_file();
    assert (res);
    delete md;
  }

}

bool wfs_storage::stat_object( wfs_file_handle fh, struct stat *sb,
			       int32_t vernum ) {
  std::string ff = filename( fh, vernum );
  if( stat( ff.c_str(), sb ) == 0 ) {
    return true;
  } else {
    return false;
  }
}

bool wfs_storage::set_stat_object( wfs_file_handle fh, struct stat *sb,
				   int32_t vernum ) {
  std::string ff = filename( fh, vernum );
  // for now, only set atime and mtime
  struct utimbuf buf;
  buf.actime = sb->st_atime;
  buf.modtime = sb->st_mtime;
  if( utime( ff.c_str(), &buf ) == 0 ) {
    return true;
  } else {
    return false;
  }
}

bool wfs_storage::object_exists( wfs_file_handle fh, int32_t vernum ) {
  struct stat sb;
  bool obj = stat_object( fh, &sb, vernum );
  if( obj ) {
    return true;
  } else {
    // see if the metadata exists -- if so, the object used to exist but
    // was deleted, and now the id shouldn't be reused
    bool is_md = false;  // for xattr, the file not existing is enough
    if( _md_memory_only ) {
      pthread_mutex_lock( &_md_lock );
      is_md = (_metadata.find(fh.fid) != _metadata.end());
      pthread_mutex_unlock( &_md_lock );
    } else if( !_md_xattr ) {
      wfs_metadata *md = get_md(fh.fid, vernum);
      is_md = (md->version.node_id != 0);
      release_md(md);
    }
    return is_md;
  }

}

bool wfs_storage::create_new_object( wfs_file_handle fh, int32_t vernum ) {

  std::string ff = filename( fh, vernum );

  // if the file already exists, stop and say so
  if( object_exists(fh) ) {
    return false;
  }

  // ~xavid: keep the physical file only accessible to us
  int fd = creat( ff.c_str(), 0600 );

  if( fd < 0 ) {
    perror(ff.c_str());
    return false;
  }
  close(fd);
  return true;

}

bool wfs_storage::delete_object( wfs_file_handle fh, int32_t vernum ) {

  std::string ff = filename( fh, vernum );

  // if the file doesn't exist, stop and say so
  if( !object_exists(fh) ) {
    return false;
  }

  wfs_log(WFS_DBG_1, "Permanently deleting object %016"PRIx64", file %s\n",
	    fh.fid, ff.c_str() );

  if( !_md_xattr ) {
    if( ::unlink( ff.c_str() ) ) {
      wfs_log(WFS_DBG_1, "Couldn't delete file %016"PRIx64", file %s\n",
          fh.fid, ff.c_str() );
      return false;
    }
  } else {
    if( ::truncate( ff.c_str(), 0 ) ) {
      wfs_log(WFS_DBG_2, "Couldn't truncate file %016"PRIx64", file %s\n",
          fh.fid, ff.c_str() );
      return false;
    }
  }

  // reduce the info in the metadata but keep it around, because otherwise
  // an old replica (with an older version) could come back to life.
  wfs_metadata *md = get_md(fh.fid);
  md->perm_cues = "";
  md->permissions.clear();
  set_and_release_md(fh.fid, md);

  return true;

}

// returns true if there was a line
bool wfs_storage::find_fid( void *dirp, const std::string &name,
			    char *line, bool use_line,
			    wfs_file_id &fid, bool &higher,
			    char **line_buf, int &line_size ) {

  FILE *fp = (FILE *) dirp;
  fid = 0;
  higher = false;
  bool matching = true;

  *line_buf = line;

  int j = 0;
  unsigned i = 0;
  while( (line[j] = fgetc(fp)) != EOF && line[j] != '\n' ) {
    if( matching && line[j] != name.c_str()[i] ) {
      matching = false;
      // didn't find it
      if( line[j] > name.c_str()[i] ) {
	higher = true;
      }
      if( !use_line ) {
	fgets( line, WFS_DIRENTRY_MAX_LEN, fp );
	return true;
      }
    }

    j++;
    i++;

    // reach the end of the name, make sure the space matches
    if( matching && i == name.length() ) {
      line[j] = fgetc(fp);
      if( line[j] == WFS_DIRENTRY_SEP ) {
	// found it;
	// SCNx32 is the macro for scanning uint32_t
	uint32_t fid_1;
	uint32_t fid_2;
	if( fscanf( fp, "%08"SCNx32"%08"SCNx32"\n", &fid_1, &fid_2 ) == 2 ) {
	  // "line" will not be used, so no need to set line_size correctly
	  fid = two_ints_to_file_id( fid_1, fid_2 );
	  line_size = 0;
	  return true;
	}
      }
      // the entry has the name as a prefix
      matching = false;
      j++;
    }

  }

  line_size = j+1;
  if( line[j] == EOF ) {
    return false;
  } else {
    j++;
    line[j] = '\0';
    return true;
  }

}

void * wfs_storage::open_finder( wfs_file_handle fh ) {
  std::string f = filename(fh);
  FILE *fp = fopen( f.c_str(), "r" );
  if( fp == 0 ) {
    wfs_log(WFS_DBG_2, "open_finder failed due to fopen!\n");
    return NULL;
  } else {
    return fp;
  }
}

void wfs_storage::close_finder( void *dirp ) {
  fclose( (FILE *) dirp );
}

bool wfs_storage::link( wfs_file_handle dh, std::string name,
			wfs_file_handle new_fh, std::string &append_data ) {

  std::string df = filename(dh);

  if( name.size() > WFS_DIRENTRY_MAX_NAMELEN ) {
    wfs_log(WFS_DBG_OFF,
	     "Tried to create file name that's too long: %s\n", name.c_str() );
    return false;
  }

  char entry[WFS_DIRENTRY_MAX_LEN];
  sprintf( entry, "%s"WFS_DIRENTRY_SEP_Q"%016"PRIx64"\n", name.c_str(),
	   file_handle_to_file_id(new_fh) );
  append_data = "";

  // don't alphabetize non-lax directories (for performance)
  if( dh.get_cue(WFS_CUE_CONSISTENCY).val != WFS_CUE_EVENTUAL ) {

    FILE *dfp = fopen( df.c_str(), "a" );
    if( dfp == 0 ) {
      perror( df.c_str() );
      return false;
    }

    fprintf( dfp, entry );

    fclose( dfp );
    append_data = std::string(entry);

  } else {

    // alphabetize lax directories

    std::string dtmp = df + ".tmp";
    void *dfp = open_finder(dh);
    if( dfp == NULL ) {
      perror( df.c_str() );
      return false;
    }

    FILE *tmp_fp = fopen( dtmp.c_str(), "w" );
    if(tmp_fp == 0){
      perror(dtmp.c_str());
      close_finder(dfp);
      return false;
    }

    // we have to add the entry in the right alphabetical order,
    // so that checksums will come out right
    wfs_file_id fid;
    bool printed = false;
    char line[WFS_DIRENTRY_MAX_LEN];
    bool higher;
    char *line_buf;
    int line_size;
    while( find_fid( dfp, name, line, true, fid, higher,
		     &line_buf, line_size ) ) {
      if( !printed && higher ) {
        fprintf( tmp_fp, entry );
        printed = true;
      }
      fprintf( tmp_fp, "%.*s", line_size, line_buf );
    }

    if( !printed ) {
      fprintf( tmp_fp, entry );
    }

    close_finder(dfp);
    fclose(tmp_fp);

    // xattr must copy explicitly to the new file
    wfs_metadata *md = NULL;
    if( _md_xattr ) {
      md = get_md(dh.fid);
    }
    ::rename( dtmp.c_str(), df.c_str() );
    if( _md_xattr ) {
      set_and_release_md( dh.fid, md );
    }

  }

  return true;

}

bool wfs_storage::lookup( wfs_file_handle dh, std::string name,
			  wfs_file_id &fid ) {

  void *fp = open_finder(dh);
  if( fp == NULL ) {
    wfs_log(WFS_DBG_2, "lookup failed due to open_finder!\n");
    return false;
  }

  wfs_file_id found_fid;
  char line[WFS_DIRENTRY_MAX_LEN];
  bool higher;
  bool lax = (dh.get_cue(WFS_CUE_CONSISTENCY).val == WFS_CUE_EVENTUAL);
  char *line_buf;
  int line_size;
  while( find_fid( fp, name, line, false, found_fid, higher,
		   &line_buf, line_size ) ) {
    if( found_fid != 0 ) {
      close_finder(fp);
      fid = found_fid;
      return true;
    }

    // alphabetical
    if( lax && higher ) {
      break;
    }

  }

  close_finder(fp);
  return true;

}

bool wfs_storage::unlink( wfs_file_handle dh, std::string name ) {

  std::string df = filename(dh);
  void *dfp = open_finder(dh);
  if( dfp == NULL ) {
    return false;
  }

  std::string dtmp = df + ".tmp";

  FILE *tmp_fp = fopen( dtmp.c_str(), "w" );
  if(tmp_fp == 0){
    perror(dtmp.c_str());
    close_finder(dfp);
    return false;
  }

  // scan through it, printing everything but the deleted entry back
  // out to disk
  //char zname[512];
  wfs_file_id fid;
  //int zind = 0;
  char line[WFS_DIRENTRY_MAX_LEN];
  bool higher; // not used
  char *line_buf;
  int line_size;
  while( find_fid( dfp, name, line, true, fid, higher,
		   &line_buf, line_size ) ) {
    if( fid == 0 ) {
      fprintf( tmp_fp, "%.*s", line_size, line_buf );
    }
  }

  close_finder(dfp);
  fclose(tmp_fp);

  // xattr must copy explicitly to the new file
  wfs_metadata *md = NULL;
  if( _md_xattr ) {
    md = get_md(dh.fid);
  }
  ::rename( dtmp.c_str(), df.c_str() );
  if( _md_xattr ) {
    set_and_release_md( dh.fid, md );
  }

  return true;

}

bool wfs_storage::readdir( wfs_file_handle fh, unsigned long long ind,
    unsigned int n, std::vector<dirent> &e ) {

  void *fp = open_finder(fh);
  if( fp == NULL ) {
    return false;
  }
  wfs_file_id fid;
  uint32_t fid_1;
  uint32_t fid_2;
  unsigned zind = 0;
  char line[WFS_DIRENTRY_MAX_LEN];
  bool higher; // not used
  char *line_buf;
  int line_size;
  while( find_fid( fp, "", line, true, fid, higher, &line_buf, line_size ) ) {

    char* end_of_name = strchr( line_buf, WFS_DIRENTRY_SEP );
    end_of_name[0] = '\0';
    if( sscanf( end_of_name+1, "%08"SCNx32"%08"SCNx32"\n",
        &fid_1, &fid_2 ) == 2 ) {
      if( zind >= ind && e.size() < n ) {
        dirent entry;
        entry.name = line_buf;
        fid = two_ints_to_file_id( fid_1, fid_2 );
        entry.inum = file_id_to_file_handle(fid);
        e.push_back(entry);
        wfs_log(WFS_DBG_1, "Returning name %s with inum %016"PRIx64
            ", fid %016"PRIx64"\n",
            line_buf, entry.inum.fid, fid );
      }
      zind++;
    }
  }
  close_finder(fp);
  return true;

}

const void wfs_storage::save_old_version( wfs_file_handle fh ) {
  wfs_metadata *md = get_md(fh.fid);
  bool resq = md->archive();
  assert (resq);
  const std::string &fn = filename(fh);
  const std::string &newn = filename(fh, md->version.version_num);
  release_md(md);
  copy_file(fn,newn);
}

bool wfs_storage::readfile( wfs_file_handle fh, wfs_version version,
			    unsigned long long off, int n, std::string &buf ) {

  std::string f = filename( fh, version.version_num );
  int fd = open(f.c_str(), O_RDONLY);
  if( fd < 0 ) {
    return false;
  }
  if( lseek( fd, off, 0 ) == -1 ) {
    close(fd);
    return false;
  }

  char *p = new char[n];
  n = read(fd, p, n);
  close(fd);

  bool success = false;
  if( n >= 0 ) {
    buf = std::string( p, n );
    success = true;
  }
  delete [] p;
  return success;

}

bool wfs_storage::writefile( wfs_file_handle fh,
			     std::vector<unsigned long long> off,
			     std::vector<std::string> buf ) {

  std::string f = filename(fh);
  int fd = open( f.c_str(), O_WRONLY );
  if( fd < 0 ) {
    return false;
  }

  assert( off.size() == buf.size() );
  for( unsigned i = 0; i < off.size(); i++ ) {

    if( lseek( fd, off[i], 0 ) == -1 ) {
      close(fd);
      return -1;
    }
    int n = write( fd, buf[i].data(), buf[i].size() );
    if( n != (int)buf[i].size() ) {
      close(fd);
      return false;
    }

  }

  if( fhsynclevel(fh) > 0 ) {
    fsync(fd);
  }

  close(fd);
  return true;

}

bool wfs_storage::truncate( wfs_file_handle fh, unsigned long long off ) {

  std::string f = filename(fh);
  int n = ::truncate( f.c_str(), off );
  return (n == 0);

}

int wfs_storage::read_from_finder( void *filep, char *buf, int size,
				   const unsigned char **blockp ) {
  FILE *fp = (FILE *) filep;
  *blockp = (unsigned char *) buf;
  return fread( buf, sizeof(char), size, fp );
}

unsigned long long wfs_storage::file_checksum( wfs_file_handle fh ) {

  int block_size = WFS_BLOCK_SIZE; //1024*1024;
  char *block = (char *) malloc( sizeof(char)*block_size );

  void *filep = open_finder(fh);
  if( filep == NULL ) {
    free(block);
    return 0;
  }

  MD5_CTX md5_context;

#if defined(__linux__) || defined(__APPLE__)
  MD5_Init( &md5_context );
#else
  MD5Init( &md5_context );
#endif

  int nread;
  do {

    const unsigned char* ucblock;
    nread = read_from_finder( filep, block, block_size, &ucblock );

    if( nread < 0 ) {
      close_finder(filep);
      free(block);
      return 0;
    }

#if defined(__linux__) || defined(__APPLE__)
    MD5_Update( &md5_context, ucblock, nread );
#else
    MD5Update( &md5_context, ucblock, nread );
#endif

  } while( nread == block_size );

  close_finder(filep);

#if defined(__linux__) || defined(__APPLE__)
  unsigned char md5_result_int[WFS_CHECKSUM_LEN/2+1];
  MD5_Final( md5_result_int, &md5_context);

  unsigned char md5_result[WFS_CHECKSUM_LEN+1];
  // have to convert the bytes manually to hex, because there's no MD5End
  char *curr_char = (char *) md5_result;
  for( uint i = 0; i < (WFS_CHECKSUM_LEN/2); i++ ) {
    sprintf( curr_char, "%02x", md5_result_int[i] );
    curr_char += 2;
  }

#else

  char md5_result[WFS_CHECKSUM_LEN+1];
  MD5End( &md5_context, md5_result );

#endif /*__linux__*/

  free(block);
  return md5_to_long( md5_result );

}
